# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 01:34:00 2021

@author: yrc2
"""
import biosteam as bst
from biosteam.utils import CABBI_colors, GG_colors, colors
from thermosteam.utils import set_figure_size, set_font, roundsigfigs
from thermosteam.units_of_measure import format_units
from colorpalette import Palette
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from warnings import warn
import numpy as np
import pandas as pd
from biorefineries import cane
import flexsolve as flx
from . import feature_mockups as features
from .feature_mockups import (
    tea_monte_carlo_metric_mockups,
    tea_monte_carlo_derivative_metric_mockups,
    lca_monte_carlo_metric_mockups, 
    lca_monte_carlo_derivative_metric_mockups,
    ROI, MFPP, TCI, electricity_production, GWP_biofuel_allocation,
    # natural_gas_consumption,
    ethanol_production, biodiesel_production, biodiesel_yield,
    GWP_ethanol, GWP_biodiesel, GWP_electricity,
    GWP_ethanol_allocation, GWP_biodiesel_allocation,
    GWP_economic, MFPP_derivative, 
    TCI_derivative, 
    ethanol_production_derivative,
    biodiesel_production_derivative,
    electricity_production_derivative,
    # natural_gas_consumption_derivative,
    GWP_ethanol_derivative,
)
from .results import (
    images_folder,
    monte_carlo_file,
    get_monte_carlo,
    get_line_monte_carlo,
    spearman_file,
)
import os
from colorpalette import ColorWheel
from.parse_configuration import format_name

__all__ = (
    'plot_all',
    # 'plot_montecarlo_main_manuscript',
    'plot_breakdowns',
    'plot_montecarlo_feedstock_comparison',
    'plot_montecarlo_configuration_comparison',
    'plot_montecarlo_agile_comparison',
    'plot_montecarlo_derivative',
    'plot_montecarlo_absolute',
    'plot_lines_monte_carlo_manuscript',
    'plot_spearman_tea',
    'plot_spearman_lca',
    'plot_spearman_tea_short',
    'plot_spearman_lca_short',
    'plot_monte_carlo_across_coordinate',
    'monte_carlo_box_plot',
    'plot_monte_carlo',
    'plot_spearman',
    'plot_configuration_breakdown',
    'plot_feedstock_conventional_comparison_kde',
    'plot_feedstock_cellulosic_comparison_kde',
    'plot_configuration_comparison_kde',
    'plot_open_comparison_kde',
    'plot_feedstock_comparison_kde',
    'plot_crude_configuration_comparison_kde',
    'plot_agile_comparison_kde',
    'plot_separated_configuration_comparison_kde',
    'plot_unlabeled_feedstock_conventional_comparison_kde',
    'plot_competitive_biomass_yield_across_oil_content',
    'plot_competitive_microbial_oil_yield_across_oil_content',
    'plot_microbial_oil_bioethanol_oilcane_comparison_kde',
    'plot_microbial_oil_bioethanol_sugarcane_comparison_kde',
    'area_colors',
    'area_hatches',
)

area_colors = {
    'Feedstock handling': CABBI_colors.teal, 
    'Juicing': CABBI_colors.green_dirty,
    'EtOH prod.': CABBI_colors.blue,
    'Ethanol production': CABBI_colors.blue,
    'Oil ext.': CABBI_colors.brown,
    'Oil extraction': CABBI_colors.brown,
    'Biod. prod.': CABBI_colors.orange,
    'Biodiesel production': CABBI_colors.orange,
    'Pretreatment': CABBI_colors.green,
    'Wastewater treatment': colors.purple,
    'CH&P': CABBI_colors.yellow,
    'Co-Heat and Power': CABBI_colors.yellow,
    'Utilities': colors.red,
    'Storage': CABBI_colors.grey,
    'HXN': colors.orange,
    'Heat exchanger network': colors.orange,
    'Oil prod. & ext.': CABBI_colors.blue,
    'Oil production and extraction': CABBI_colors.blue,
}

area_hatches = {
    'Feedstock handling': 'x', 
    'Juicing': '-',
    'EtOH prod.': '/',
    'Ethanol production': '/',
    'Oil ext.': '\\',
    'Oil extraction': '\\',
    'Biod. prod.': '/|',
    'Biodiesel production': '/|',
    'Pretreatment': '//',
    'Wastewater treatment': r'\\',
    'CH&P': '',
    'Co-Heat and Power': '',
    'Utilities': '\\|',
    'Storage': '',
    'HXN': '+',
    'Heat exchanger network': '+',
    'TAG prod.': '/',
    'Oil prod. & ext.': '/',
    'Oil production and extraction': '/',
}
line_color_wheel = ColorWheel([
    GG_colors.orange,
    GG_colors.purple,
    GG_colors.green,
    GG_colors.blue,
    GG_colors.yellow,
    colors.CABBI_teal,
    colors.CABBI_grey,
    colors.CABBI_brown,
])
for i, color in enumerate(line_color_wheel.colors): line_color_wheel.colors[i] = color.tint(20)
edgecolor = (*colors.CABBI_black.RGBn, 1)
for i in area_colors: area_colors[i] = area_colors[i].tint(20)
palette = Palette(**area_colors)
letter_color = colors.neutral.shade(25).RGBn
GWP_units_L = '$\\mathrm{kg} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{L}^{-1}$'
GWP_units_GGE = GWP_units_L.replace('L', 'GGE')
GWP_units_L_small = GWP_units_L.replace('kg', 'g')
CABBI_colors.orange_hatch = CABBI_colors.orange.copy(hatch='////')
    
ethanol_over_biodiesel = bst.MockVariable('Ethanol over biodiesel', 'L/MT', 'Biorefinery')
GWP_ethanol_displacement = features.GWP_ethanol_displacement
production = (ethanol_production, biodiesel_production)

mc_metric_settings = {
    'MFPP': (MFPP, f"MFPP\n[{format_units('USD/MT')}]", None),
    'TCI': (TCI, f"TCI\n[{format_units('10^6*USD')}]", None),
    'production': (production, f"Production\n[{format_units('L/MT')}]", None),
    'electricity_production': (electricity_production, f"Elec. prod.\n[{format_units('kWhr/MT')}]", None),
    # 'natural_gas_consumption': (natural_gas_consumption, f"NG cons.\n[{format_units('m^3/MT')}]", None),
    'GWP_ethanol_displacement': (GWP_ethanol_displacement, "GWP$_{\\mathrm{displacement}}$" f"\n[{GWP_units_L}]", None),
    'GWP_economic': ((GWP_ethanol, GWP_biodiesel), "GWP$_{\\mathrm{economic}}$" f"\n[{GWP_units_L}]", None),
    'GWP_energy': ((GWP_ethanol_allocation, GWP_biodiesel_allocation), "GWP$_{\\mathrm{energy}}$" f"\n[{GWP_units_L}]", None),
}

mc_line_metric_settings = {
    'Biodiesel production': (biodiesel_production, f"Biodiesel production\n[{format_units('L/MT')}]"),
    'Biodiesel yield': (biodiesel_yield, f"Biodiesel yield\n[{format_units('L/ha')}]"),
    'GWP biodiesel': (GWP_biodiesel, f"GWP Biodiesel [{GWP_units_L}]"),
    'GWP biofuel': (GWP_biofuel_allocation, f"GWP [{GWP_units_GGE}]"),
    'ROI': (ROI, "ROI [%]"),
}

mc_comparison_settings = {
    'MFPP': (MFPP, r"$\Delta$" + f"MFPP\n[{format_units('USD/MT')}]", None),
    'TCI': (TCI, r"$\Delta$" + f"TCI\n[{format_units('10^6*USD')}]", None),
    'production': (production, r"$\Delta$" + f"Production\n[{format_units('L/MT')}]", None),
    'electricity_production': (electricity_production, r"$\Delta$" + f"Elec. prod.\n[{format_units('kWhr/MT')}]", None),
    # 'natural_gas_consumption': (natural_gas_consumption, r"$\Delta$" + f"NG cons.\n[{format_units('m^3/MT')}]", None),
    'GWP_ethanol_displacement': (GWP_ethanol_displacement, r"$\Delta$" + "GWP$_{\\mathrm{displacement}}$" f"\n[{GWP_units_L}]", None),
    'GWP_economic': (GWP_ethanol, r"$\Delta$" + "GWP$_{\\mathrm{economic}}$" f"\n[{GWP_units_L}]", None),
    'GWP_energy': (GWP_ethanol_allocation, r"$\Delta$" + "GWP$_{\\mathrm{energy}}$" f"\n[{GWP_units_L}]", None),
    'GWP_property_allocation': ((GWP_ethanol, GWP_ethanol_allocation), r"$\Delta$" + f"GWP\n[{GWP_units_L}]", None),
}

mc_derivative_metric_settings = {
    'MFPP': (MFPP_derivative, r"$\Delta$" + format_units(r"MFPP/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('USD/MT')}]", None),
    'TCI': (TCI_derivative,  r"$\Delta$" + format_units(r"TCI/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('10^6*USD')}]", None),
    'production': ((ethanol_production_derivative, biodiesel_production_derivative), r"$\Delta$" + format_units(r"Prod./OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('L/MT')}]", None),
    'electricity_production': (electricity_production_derivative, r"$\Delta$" + format_units(r"EP/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('kWhr/MT')}]", None),
    # 'natural_gas_consumption': (natural_gas_consumption_derivative, r"$\Delta$" + format_units(r"NGC/OC").replace('cdot', r'cdot \Delta') + f"\n[{format_units('m^3/MT')}]", None),
    'GWP_economic': (GWP_ethanol_derivative, r"$\Delta$" + r"GWP $\cdot \Delta \mathrm{OC}^{-1}$" f"\n[{GWP_units_L_small}]", 1000),
}

kde_metric_settings = {j[0]: j for j in mc_metric_settings.values()}
kde_comparison_settings = {j[0]: j for j in mc_comparison_settings.values()}
kde_derivative_settings = {j[0]: j for j in mc_derivative_metric_settings.values()}

# %% Plots for publication

def plot_all():
    # plot_montecarlo_main_manuscript()
    plot_montecarlo_absolute()
    plot_spearman_tea()
    plot_spearman_lca()
    plot_breakdowns()

def plot_montecarlo_feedstock_comparison(axes_box=None, letters=None, 
                                         single_column=True):
    if single_column:
        width = 'half'
        aspect_ratio = 2.25
        ncols = 1
        left = 0.255
        bottom = 0.05
    else:
        width = None
        aspect_ratio = 0.75
        left = 0.105
        bottom = 0.12
        ncols = 3
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=width, aspect_ratio=aspect_ratio)       
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile=False, ncols=ncols, axes_box=axes_box,
        labels=[
            'Direct Cogeneration',
            'Integrated Co-Fermentation',
            # 'Direct Cogeneration',
            # 'Integrated Co-Fermentation',
        ],
        comparison_names=['O1 - S1', 'O2 - S2'],
        metrics = ['MFPP', 'TCI', 'production', 'GWP_property_allocation', 
                   'natural_gas_consumption', 'electricity_production'],
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 
            'orange', 'orange_hatch', 'grey', 'brown',
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGH' if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
        # if axes_box is None and letter in 'DH':
        #     x = 0.5
        #     plt.text(x, ylb - (yub - ylb) * 0.3, 
        #              'Impact of processing\noilcane over sugarcane', 
        #              horizontalalignment='center',verticalalignment='center',
        #              fontsize=8)
    if axes_box is None:
        plt.subplots_adjust(right=0.96, left=left, wspace=0.38, top=0.98, bottom=bottom)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_feedstock_comparison.{i}')
            plt.savefig(file, transparent=True)

    
def plot_montecarlo_configuration_comparison(axes_box=None, letters=None,
                                             single_column=True):
    if single_column:
        width = 'half'
        aspect_ratio = 2.25
        ncols = 1
        left = 0.255
        bottom = 0.05
        x = 1.65
        metrics= ['MFPP', 'TCI', 'production', 'GWP_property_allocation',
                  'natural_gas_consumption', 'electricity_production']
    else:
        width = None
        aspect_ratio = 0.75
        left = 0.105
        bottom = 0.12
        ncols = 2
        x = 0.58
        metrics= ['MFPP', 'TCI', 'production', 'GWP_property_allocation']
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=width, aspect_ratio=aspect_ratio)
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile=False, ncols=ncols, axes_box=axes_box,
        labels=[
            'Oilcane',
            # 'Sugarcane',
        ],
        comparison_names=[
            'O2 - O1', 
            # 'S2 - S1'
        ],
        metrics=metrics,
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 
            'orange', 'orange_hatch', 
        ])
    )
    for ax, letter in zip(axes, 'ABCDEF' if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(x, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    if axes_box is None:
        plt.subplots_adjust(right=0.96, left=left, wspace=0.38, top=0.98, bottom=bottom)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_configuration_comparison.{i}')
            plt.savefig(file, transparent=True)
    
def plot_montecarlo_agile_comparison(axes_box=None, letters=None):
    if axes_box is None:
        set_font(size=8)
        set_figure_size(width=3.3071, aspect_ratio=1.0)
    fig, axes = plot_monte_carlo(
        derivative=False, absolute=False, comparison=True,
        tickmarks=None, agile_only=True, ncols=1,
        labels=[
            'Direct Cogeneration',
            'Integrated Co-Fermentation'
        ],
        metrics=['MFPP', 'TCI'],
        axes_box=axes_box,
    )
    for ax, letter in zip(axes, 'AB'  if letters is None else letters):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
        if axes_box is None and letter == 'B':
            plt.text(0.5, ylb - (yub - ylb) * 0.25, 
                      'Impact of integrating oilsorghum\nat an agile oilcane biorefinery', 
                      horizontalalignment='center',verticalalignment='center',
                      fontsize=8)
    if axes_box is None:
        plt.subplots_adjust(right=0.9, left=0.2, wspace=0.5, top=0.98, bottom=0.15)
        for i in ('svg', 'png'):
            file = os.path.join(images_folder, f'montecarlo_agile_comparison.{i}')
            plt.savefig(file, transparent=True)
    
def plot_montecarlo_derivative():
    set_font(size=8)
    set_figure_size(
        aspect_ratio=0.5,
        # width=3.3071, aspect_ratio=1.85
    )
    fig, axes = plot_monte_carlo(
        derivative=True, absolute=True, 
        comparison=False, agile=False,
        ncols=3,
        # tickmarks=np.array([
        #     [-3, -2, -1, 0, 1, 2, 3, 4, 5],
        #     [-9, -6, -3,  0, 3, 6, 9, 12, 15],
        #     [-2.0, -1.5, -1.0, -0.5, 0, 0.5, 1.0, 1.5, 2],
        #     [-16, -8, 0, 8, 16, 24, 32, 40, 48],
        #     [-400, -300, -200, -100, 0, 100, 200, 300, 400],
        #     [-300, -225, -150, -75, 0, 75, 150, 225, 300]
        # ], dtype=object),
        labels=['DC', 'ICF'],
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 'grey', 
            'orange', 
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGH'):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(1.65, ylb + (yub - ylb) * 0.90, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    plt.subplots_adjust(
        hspace=0, wspace=0.7,
        top=0.95, bottom=0.1,
        left=0.12, right=0.96
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_derivative.{i}')
        plt.savefig(file, transparent=True)

def plot_montecarlo_absolute():
    set_font(size=8)
    set_figure_size(aspect_ratio=1.05)
    fig, axes = plot_monte_carlo(
        absolute=True, comparison=False, ncols=2,
        expand=0.1, 
        labels=['Sugarcane\nDC', 'Oilcane\nDC',
                'Sugarcane\nICF', 'Oilcane\nICF',
                'Sugarcane &\nSorghum DC', 'Oilcane &\nOil-sorghum DC',
                'Sugarcane &\nSorghum ICF', 'Oilcane &\nOil-sorghum ICF'],
        xrot=90,
        color_wheel = CABBI_colors.wheel([
            'blue_light', 'green_dirty', 'orange', 'green', 'grey',
            'orange', 'orange', 'green', 'orange', 'green',
        ])
    )
    for ax, letter in zip(axes, 'ABCDEFGHIJ'):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(7.8, ylb + (yub - ylb) * 0.92, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    plt.subplots_adjust(left=0.12, right=0.95, wspace=0.40, top=0.98, bottom=0.2)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_absolute.{i}')
        plt.savefig(file, transparent=True)
    
def plot_lines_monte_carlo_manuscript(fs=10):
    set_font(size=fs)
    set_figure_size(aspect_ratio=.65)
    fig, axes = plot_lines_monte_carlo(
        ncols=1,
        metrics=['ROI', 'GWP biofuel'],
        labels=['Sugarcane', 'Energycane', 
                '1566\n1.8% Oil', 
                '1565\n2.0% Oil',
                '1580\n5.4% Oil'],
        expand=0.1, 
        xrot=45,
        color_wheel = line_color_wheel,
    )
    for ax, letter in zip(axes.flat, 'ABCDEFGHIJ'):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        plt.text(-0.25, ylb + (yub - ylb) * 0.92, letter, color=letter_color,
                 horizontalalignment='center',verticalalignment='center',
                 fontsize=12, fontweight='bold')
    plt.subplots_adjust(left=0.12, right=0.95, wspace=0, top=0.95, bottom=0.2)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'montecarlo_lines.{i}')
        plt.savefig(file, transparent=True)
    
def plot_spearman_tea(with_units=None, aspect_ratio=0.8, **kwargs):
    set_font(size=8)
    set_figure_size(aspect_ratio=aspect_ratio)
    plot_spearman(
        configurations=[
            'O1', 'O1*',
            'O2', 'O2*',
        ],
        labels=[
            'DC', 'Oil-sorghum int., DC',
            'ICF', 'Oil-sorghum int., ICF',
        ],
        kind='TEA',
        with_units=with_units,
        cutoff=0.05,
        **kwargs
    )
    plt.subplots_adjust(left=0.45, right=0.975, top=0.98, bottom=0.08)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_tea.{i}')
        plt.savefig(file, transparent=True)

def plot_spearman_tea_short(**kwargs):
    set_font(size=8)
    set_figure_size(aspect_ratio=0.65, width=6.6142 * 2/3)
    plot_spearman(
        configurations=[
            'O1', 
            'O2', 
        ],
        labels=[
            'DC', 
            'ICF',
        ],
        kind='TEA',
        with_units=False,
        cutoff=0.05,
        top=5,
        legend=True,
        legend_kwargs={'loc': 'upper left'},
        **kwargs
    )
    plt.subplots_adjust(left=0.35, right=0.975, top=0.98, bottom=0.15)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_tea.{i}')
        plt.savefig(file, transparent=True)

def plot_spearman_lca_short(with_units=False, aspect_ratio=0.65, **kwargs):
    set_font(size=8)
    set_figure_size(aspect_ratio=aspect_ratio, width=6.6142 * 2/3)
    plot_spearman(
        configurations=[
            'O1', 
            'O2', 
        ],
        labels=[
            'DC', 
            'ICF',
        ],
        kind='LCA',
        with_units=with_units,
        cutoff=0.025,
        top=5,
        legend=False,
        **kwargs
    )
    plt.subplots_adjust(left=0.35, right=0.975, top=0.98, bottom=0.15)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_lca.{i}')
        plt.savefig(file, transparent=True)

def plot_spearman_lca(with_units=None, aspect_ratio=0.65, **kwargs):
    set_font(size=8)
    set_figure_size(aspect_ratio=aspect_ratio)
    plot_spearman(
        configurations=[
            'O1', 'O1*',
            'O2', 'O2*',
        ],
        labels=[
            'DC', 'Oil-sorghum int., DC',
            'ICF', 'Oil-sorghum int., ICF',
        ],
        kind='LCA',
        with_units=with_units,
        cutoff=0.025,
        **kwargs
    )
    plt.subplots_adjust(left=0.45, right=0.975, top=0.98, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'spearman_lca.{i}')
        plt.savefig(file, transparent=True)

def plot_breakdowns(biodiesel_only=False):
    set_font(size=8)
    set_figure_size(aspect_ratio=0.68)
    fig, axes = plt.subplots(nrows=1, ncols=2)
    plt.sca(axes[0])
    c1, c2 = ('O7', 'O8') if biodiesel_only else ('O1', 'O2')
    plot_configuration_breakdown(c1, ax=axes[0], legend=False)
    plt.sca(axes[1])
    plot_configuration_breakdown(c2, ax=axes[1], legend=True)
    yticks = axes[1].get_yticks()
    plt.yticks(yticks, ['']*len(yticks))
    plt.ylabel('')
    plt.subplots_adjust(left=0.09, right=0.96, wspace=0., top=0.84, bottom=0.31)
    for ax, letter in zip(axes, ['(A) Direct Cogeneration', '(B) Integrated Co-Fermentation']):
        plt.sca(ax)
        ylb, yub = plt.ylim()
        xlb, xub = plt.xlim()
        plt.text((xlb + xub) * 0.5, ylb + (yub - ylb) * 1.2, letter, color=letter_color,
                  horizontalalignment='center',verticalalignment='center',
                  fontsize=12, fontweight='bold')
    for i in ('svg', 'png'):
        name = f'breakdowns_biodiesel_only.{i}' if biodiesel_only else f'breakdowns.{i}'
        file = os.path.join(images_folder, name)
        plt.savefig(file, transparent=True)

# %% KDE


def plot_kde(name, metrics=(GWP_ethanol, MFPP), xticks=None, yticks=None,
             xbox_kwargs=None, ybox_kwargs=None, top_left='',
             top_right='Tradeoff', bottom_left='Tradeoff',
             bottom_right='', fs=None, ticklabels=True, aspect_ratio=1.1,
             line=None):
    set_font(size=fs or 8)
    set_figure_size(width='half', aspect_ratio=aspect_ratio)
    Xi, Yi = [i.index for i in metrics]
    df = get_monte_carlo(name, metrics, line)
    y = df[Yi].values
    x = df[Xi].values
    sX, sY = [kde_comparison_settings[i] for i in metrics]
    _, xlabel, fx = sX
    _, ylabel, fy = sY
    if fx: x *= fx
    if fy: y *= fy
    ax = bst.plots.plot_kde(
        y=y, x=x, xticks=xticks, yticks=yticks,
        xticklabels=ticklabels, yticklabels=ticklabels,
        xbox_kwargs=xbox_kwargs or dict(light=CABBI_colors.orange.RGBn, dark=CABBI_colors.orange.shade(60).RGBn),
        ybox_kwargs=ybox_kwargs or dict(light=CABBI_colors.blue.RGBn, dark=CABBI_colors.blue.shade(60).RGBn),
        aspect_ratio=1.2,
    )
    plt.sca(ax)
    plt.xlabel(xlabel.replace('\n', ' '))
    plt.ylabel(ylabel.replace('\n', ' '))
    bst.plots.plot_quadrants(data=[x, y], text=[top_left, top_right, bottom_left, bottom_right])
    plt.subplots_adjust(
        hspace=0.05, wspace=0.05,
        top=0.98, bottom=0.15,
        left=0.2, right=0.98,
    )

def plot_kde_fake_scenarios_ethanol_price(name, xticks=None, yticks=None,
             xbox_kwargs=None, ybox_kwargs=None, top_left='',
             top_right='Tradeoff', bottom_left='Tradeoff',
             bottom_right='', fs=None, ticklabels=True, aspect_ratio=1.1):
    set_font(size=fs or 8)
    set_figure_size(width='half', aspect_ratio=aspect_ratio)
    ethanol_price = features.set_ethanol_price
    relative_biodiesel_price = features.set_biodiesel_price
    metrics = (GWP_ethanol, MFPP)
    all_features = (*metrics, ethanol_price, relative_biodiesel_price, 
                    ethanol_production, biodiesel_production)
    Xi, Yi = [i.index for i in metrics]
    df = get_monte_carlo(name, all_features)
    x = df[Xi].values
    y = df[Yi].values
    ethanol_price_range = [0.269, 0.758 * 2]
    biodiesel_price_range = [0.0819, 1.09 * 2]
    settings = kde_metric_settings if '-' in name else  kde_comparison_settings
    sX, sY = [settings[i] for i in metrics]
    _, xlabel, fx = sX
    _, ylabel, fy = sY
    if fx: x *= fx
    if fy: y *= fy
    xs = (x, x)
    ys = tuple([y + (i - df[ethanol_price.index]) * df[ethanol_production.index]
                  + (i + j - df[ethanol_price.index] + df[relative_biodiesel_price.index]) * df[biodiesel_production.index]
                for (i, j) in zip(ethanol_price_range, biodiesel_price_range)])
    
    print(x.min(), x.max())
    print(min(ys[0].min(), ys[1].min()), max(ys[0].max(), ys[1].max()))
    ax = bst.plots.plot_kde(
        y=ys, x=xs, xticks=xticks, yticks=yticks,
        xticklabels=ticklabels, yticklabels=ticklabels,
        xbox_kwargs=xbox_kwargs or dict(light=CABBI_colors.orange.RGBn, dark=CABBI_colors.orange.shade(60).RGBn),
        ybox_kwargs=ybox_kwargs or dict(light=CABBI_colors.blue.RGBn, dark=CABBI_colors.blue.shade(60).RGBn),
        cmaps=['inferno', 'viridis'],
        aspect_ratio=1.2,
    )
    plt.sca(ax)
    plt.xlabel(xlabel.replace('\n', ' '))
    plt.ylabel(ylabel.replace('\n', ' '))
    bst.plots.plot_quadrants(
        data=None, 
        text=['Market\nCompetitive', 'Market\nCompetitive',
              'Financial\nRisk', 'Financial\nRisk'],
        y=50,
    )
    plt.subplots_adjust(
        hspace=0.05, wspace=0.05,
        top=0.98, bottom=0.15,
        left=0.15, right=0.98,
    )

def plot_kde_2d(name, metrics=(GWP_ethanol, MFPP), xticks=None, yticks=None,
                top_left='', top_right='Tradeoff', bottom_left='Tradeoff',
                bottom_right='', xbox_kwargs=None, ybox_kwargs=None, titles=None,
                fs=None, ticklabels=True, line=None):
    set_font(size=fs or 8)
    set_figure_size(aspect_ratio=0.6)
    if isinstance(name, str): name = (name,)
    if isinstance(line, str): line = len(name)*(line,)
    Xi, Yi = [i.index for i in metrics]
    dfs = [get_monte_carlo(i, metrics, line) for i in zip(name, line)]
    sX, sY = [kde_comparison_settings[i] for i in metrics]
    _, xlabel, fx = sX
    _, ylabel, fy = sY
    xs = np.array([[df[Xi].values for df in dfs]])
    ys = np.array([[df[Yi].values for df in dfs]])
    if fx: xs *= fx
    if fy: ys *= fy
    ticklabels = [True, True] if ticklabels else [False, False]
    axes = bst.plots.plot_kde_2d(
        xs=xs, ys=ys,
        xticks=xticks, yticks=yticks,
        xticklabels=ticklabels, yticklabels=ticklabels,
        xbox_kwargs=2*[xbox_kwargs or dict(light=CABBI_colors.orange.RGBn, dark=CABBI_colors.orange.shade(60).RGBn)],
        ybox_kwargs=[ybox_kwargs or dict(light=CABBI_colors.blue.RGBn, dark=CABBI_colors.blue.shade(60).RGBn)],
        aspect_ratio=1.,
    )
    M, N = axes.shape
    text = [top_left, top_right, bottom_left, bottom_right]
    for i in range(M):
        for j in range(N):
            ax = axes[i, j]
            plt.sca(ax)
            if i == M - 1: plt.xlabel(xlabel.replace('\n', ' '))
            if j == 0: plt.ylabel(ylabel.replace('\n', ' '))
            df = dfs[j]
            x = df[Xi]
            y = df[Yi]
            bst.plots.plot_quadrants(data=[x, y], text=text)
    plt.subplots_adjust(
        hspace=0, wspace=0,
        top=0.98, bottom=0.15,
        left=0.1, right=0.98,
    )
    if titles:
        plt.subplots_adjust(
            top=0.90,
        )
        for ax, letter in zip(axes[0, :], titles):
            plt.sca(ax)
            ylb, yub = plt.ylim()
            xlb, xub = plt.xlim()
            plt.text((xlb + xub) * 0.5, ylb + (yub - ylb) * 1.17, letter, color=letter_color,
                      horizontalalignment='center', verticalalignment='center',
                      fontsize=12, fontweight='bold')

def plot_unlabeled_feedstock_conventional_comparison_kde(fake_scenarios=True):
    mfpp = kde_comparison_settings[MFPP]
    gwp = kde_comparison_settings[GWP_ethanol]
    gwp_dummy = gwp[1].replace('$_{\\mathrm{economic}}$', '')
    if fake_scenarios: 
        gwp_dummy = gwp_dummy.replace(r"$\Delta$", '')
    kde_comparison_settings[GWP_ethanol] = [
        gwp[0], 
        gwp_dummy, 
        gwp[2]
    ]
    kde_comparison_settings[MFPP] = [
        mfpp[0], 
        'IRR [%]', 
        mfpp[2]
    ]
    (plot_kde_fake_scenarios_ethanol_price if fake_scenarios else plot_kde)(
        'O1' if fake_scenarios else 'O1 - S1',
        yticks=[-75, 0, 75, 150, 225, 300] if fake_scenarios else [-15, 0, 15, 30, 45, 60],
        xticks=[0.25, 0.3, 0.35, 0.4] if fake_scenarios else [-0.12, -0.09, -0.06, -0.03, 0, 0.03, 0.06],
        top_left='Conf. A\nFavored()',
        bottom_right='Conf. B\nFavored()',
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        ticklabels=False,
        aspect_ratio=0.8,
    )
    kde_comparison_settings[GWP_ethanol] = gwp
    kde_comparison_settings[MFPP] = mfpp
    for i in ('svg', 'png'):
        file = os.path.join(
            images_folder, 
            f'prelim_scenario_comparison_kde.{i}' if fake_scenarios else f'prelim_comparison_kde.{i}'
        )
        plt.savefig(file, transparent=True)

def plot_feedstock_conventional_comparison_kde():
    plot_kde(
        'O1 - S1',
        yticks=[-15, 0, 15, 30, 45, 60],
        xticks=[-0.12, -0.09, -0.06, -0.03, 0, 0.03, 0.06],
        top_left='Oilcane Favored',
        bottom_right='Sugarcane\nFavored',
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'feedstock_conventional_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_feedstock_cellulosic_comparison_kde(fs=None):
    plot_kde(
        'O2 - S2',
        yticks=[-15, 0, 15, 30, 45, 60],
        xticks=[-0.2, -0.1, 0, 0.1, 0.2],
        top_left='Oilcane\nFavored()',
        bottom_right='Sugarcane\nFavored()',
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        fs=fs,
        # fx=1000.,
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'feedstock_cellulosic_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_feedstock_comparison_kde(fs=None):
    plot_kde_2d(
        ('O1 - S1', 'O2 - S2'),
        yticks=[[-18, 0, 18, 36, 54]],
        xticks=[[-0.12, -0.09, -0.06, -0.03, 0, 0.03, 0.06],
                [-0.06, -0.03, 0., 0.03, 0.06, 0.09]],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Oilcane\nFavored()',
        bottom_right='\nSugarcane\nFavored()',
        titles=['(A) Direct Cogeneration', '(B) Integrated Co-Fermentation'],
        fs=fs,
    )
    plt.subplots_adjust(
        wspace=0,
        
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'feedstock_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_configuration_comparison_kde(fs=None):
    plot_kde(
        'O2 - O1',
        yticks=sorted([-1 * i for i in [-39, -26, -13, 0, 13, 26]]),
        xticks=sorted([-1 * i for i in [-0.06, -0.04, -0.02, 0, 0.02, 0.04]]),
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='ICF\nFavored()',
        bottom_right='DC Favored()',
        fs=fs,
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'configuration_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_microbial_oil_bioethanol_oilcane_comparison_kde(fs=None):
    plot_kde(
        'O8 - O2',
        # yticks=sorted([-1 * i for i in [-39, -26, -13, 0, 13, 26]]),
        # xticks=sorted([-1 * i for i in [-0.06, -0.04, -0.02, 0, 0.02, 0.04]]),
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Microbial Oil\nFavored()',
        bottom_right='Bioethanol\nFavored()',
        fs=fs,
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'microbial_oil_bioethanol_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_microbial_oil_bioethanol_sugarcane_comparison_kde(fs=None):
    plot_kde(
        'O7 - S1',
        # yticks=sorted([-1 * i for i in [-39, -26, -13, 0, 13, 26]]),
        # xticks=sorted([-1 * i for i in [-0.06, -0.04, -0.02, 0, 0.02, 0.04]]),
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Microbial Oil\nFavored()',
        bottom_right='Bioethanol\nFavored()',
        line='WT',
        fs=fs,
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'microbial_oil_bioethanol_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_separated_configuration_comparison_kde():
    plot_kde_2d(
        ('O1', 'O2'),
        yticks=[[-20, 0, 20, 40, 60]],
        xticks=[
            [0, 0.5, 1, 1.5],
            [0, 2, 4, 6, 8, 10]
        ],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='DC Favored()',
        bottom_right='ICF\nFavored()',
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'separated_configuration_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_crude_configuration_comparison_kde():
    plot_kde_2d(
        ('O1 - O3', 'O2 - O4'),
        yticks=[[0, 10, 20, 30, 40]],
        xticks=[
            [-0.3, -0.24, -0.18, -0.12, -0.06, 0],
            [-0.15, -0.12, -0.09, -0.06, -0.03, 0]
        ],
        top_right='GWP\nTradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Biodiesel\nProduction Favored()',
        bottom_right='Crude Oil\nProduction Favored()',
        titles=['(A) Direct Cogeneration', '(B) Integrated Co-Fermentation'],
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'crude_configuration_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_agile_comparison_kde(fs=None):
    plot_kde_2d(
        ('O1* - O1', 'O2* - O2'),
        metrics=[TCI, MFPP],
        yticks=[[0, 3, 6, 9, 12, 15]], 
        xticks=[
            [-100, -80, -60, -40, -20, 0],
            [-150, -125, -100, -75, -50, -25, 0],
        ],
        top_right='TCI-Tradeoff()',
        bottom_left='MFPP\nTradeoff()',
        top_left='Oil-sorghum\nIntegration Favored()',
        bottom_right='Cane-only\nFavored()',
        xbox_kwargs=dict(light=CABBI_colors.green_dirty.RGBn, 
                         dark=CABBI_colors.green_dirty.shade(60).RGBn),
        titles=['(A) Direct Cogeneration', '(B) Integrated Co-Fermentation'],
        fs=fs,
    )
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'agile_conventional_comparison_kde.{i}')
        plt.savefig(file, transparent=True)

def plot_open_comparison_kde(overlap=False):
    metrics = [MFPP, TCI, GWP_ethanol, biodiesel_production]
    df_conventional_oc = get_monte_carlo('O1', metrics)
    df_cellulosic_oc = get_monte_carlo('O2', metrics)
    df_conventional_sc = get_monte_carlo('S1', metrics)
    df_cellulosic_sc = get_monte_carlo('S2', metrics)
    MFPPi = MFPP.index
    TCIi = TCI.index
    if overlap:
        ys = np.zeros([1, 2], dtype=object)
        xs = np.zeros([1, 2], dtype=object)    
        ys[0, 0] = (df_conventional_oc[MFPPi], df_cellulosic_oc[MFPPi])
        ys[0, 1] = (df_conventional_sc[MFPPi], df_cellulosic_sc[MFPPi])
        xs[0, 0] = (df_conventional_oc[TCIi], df_cellulosic_oc[TCIi])
        xs[0, 1] = (df_conventional_sc[TCIi], df_cellulosic_sc[TCIi])
        yticks = [[-30, -15, 0, 15, 30, 45, 60, 75]]
        xticks = 2*[[200, 300, 400, 500, 600]]
    else:
        ys = np.array([
            [df_conventional_oc[MFPPi], df_conventional_sc[MFPPi]],
            [df_cellulosic_oc[MFPPi], df_cellulosic_sc[MFPPi]]
        ])
        xs = np.array([
            [df_conventional_oc[TCIi], df_conventional_sc[TCIi]],
            [df_cellulosic_oc[TCIi], df_cellulosic_sc[TCIi]]
        ])
        yticks = 2*[[-30, -15, 0, 15, 30, 45, 60, 75]]
        xticks = 2*[[200, 300, 400, 500, 600]]
    bst.plots.plot_kde_2d(
        ys=ys, xs=xs, xticks=xticks, yticks=yticks,
        xbox_kwargs=[dict(position=1), dict(position=1)],
        ybox_kwargs=[dict(position=0), dict(position=0)],
    )

#%%  General Monte Carlo box plots

def plot_monte_carlo_across_coordinate(coordinate, data, color_wheel):
    if isinstance(data, list):
        return [plot_monte_carlo_across_coordinate(coordinate, i, color_wheel) for i in data]
    else:
        color = color_wheel.next()
        return bst.plots.plot_montecarlo_across_coordinate(
            coordinate, data,
            light_color=color.tint(50).RGBn,
            dark_color=color.shade(50).RGBn,
        )

def monte_carlo_box_plot(data, positions, light_color, dark_color, width=None, 
                         hatch=None, outliers=False, **kwargs):
    if width is None: width = 0.8
    if outliers:
        flierprops = {'marker':'D',
                  'markerfacecolor': light_color,
                  'markeredgecolor': dark_color,
                  'markersize':3}
    else:
        flierprops = {'marker':''}
    bp = plt.boxplot(
        x=data, positions=positions, patch_artist=True,
        widths=width, whis=[5, 95],
        boxprops={'facecolor':light_color,
                  'edgecolor':dark_color},
        medianprops={'color':dark_color,
                     'linewidth':1.5},
        flierprops=flierprops,
        **kwargs
    )
    if hatch:
        for box in bp['boxes']:
            box.set(hatch = hatch)

def plot_monte_carlo(derivative=False, absolute=True, comparison=True,
                     configuration_names=None, comparison_names=None,
                     metrics=None, labels=None, tickmarks=None, agile=True, 
                     ncols=1, expand=None, step_min=None,
                     agile_only=False, xrot=None,
                     color_wheel=None, axes_box=None):
    if derivative:
        default_configuration_names = ['O1', 'O2']
        default_comparison_names = ['O2 - O1']
        metric_info = mc_derivative_metric_settings
        default_metrics = list(metric_info)
    else:
        default_configuration_names = configuration_names[:-2]
        default_comparison_names = comparison_names
        if comparison:
            metric_info = mc_comparison_settings
        else:
            metric_info = mc_metric_settings
        if agile_only:
            default_configuration_names = [i for i in default_configuration_names if '*' in i]
            default_comparison_names = [i for i in default_comparison_names if '*' in i]
            default_metrics = ['MFPP', 'TCI', 'production']
        else:
            default_metrics = list(metric_info)
    if configuration_names is None: configuration_names = default_configuration_names
    if comparison_names is None: comparison_names = default_comparison_names
    if metrics is None: metrics = default_metrics
    combined = absolute and comparison
    if agile_only:
        configuration_names = [i for i in configuration_names if '*' in i]
        comparison_names = [i for i in comparison_names if '*' in i]
    elif not agile:
        configuration_names = [i for i in configuration_names if '*' not in i]
        comparison_names = [i for i in comparison_names if '*' not in i]
    if combined:
        columns = configurations = configuration_names + comparison_names
    elif absolute:
        columns = configurations = configuration_names
    elif comparison:
        columns = configurations = comparison_names
    else:
        columns = configurations = []
    rows, ylabels, factors = zip(*[metric_info[i] for i in metrics])
    factors = [(i, j) for i, j in enumerate(factors) if j is not None]
    if color_wheel is None: color_wheel = CABBI_colors.wheel()
    N_rows = len(rows)
    if axes_box is None:
        nrows = int(round(N_rows / ncols))
        fig, axes_box = plt.subplots(ncols=ncols, nrows=nrows)
        plt.subplots_adjust(wspace=0.45)
    else:
        nrows = N_rows
        fig = None
    axes = axes_box.transpose()
    axes = axes.flatten()
    N_cols = len(columns)    
    xtext = labels or [format_name(i).replace(' ', '') for i in configurations]
    N_marks = len(xtext)
    xticks = tuple(range(N_marks))
    
    def get_data(metric, name):
        try:
            df = get_monte_carlo(name, metric)
        except:
            return np.zeros([1, 1])
        else:
            values = df.values
            return values
    
    def plot(arr, position):
        if arr.ndim == 2:
            N = arr.shape[1]
            width = 0.618 / N
            boxwidth = 0.618 / (N + 1/N)
            plots = []
            for i in range(N):
                color = color_wheel.next()
                boxplot = monte_carlo_box_plot(
                    data=arr[:, i], positions=[position + (i-(N-1)/2)*width], 
                    light_color=color.RGBn, 
                    dark_color=color.shade(60).RGBn,
                    width=boxwidth,
                    hatch=getattr(color, 'hatch', None),
                )
                plots.append(boxplot)
            return plots
        else:
            color = color_wheel.next()
            return monte_carlo_box_plot(
                data=arr, positions=[position], 
                light_color=color.RGBn, 
                dark_color=color.shade(60).RGBn,
                width=0.618,
            )
    
    data = np.zeros([N_rows, N_cols], dtype=object)
    data[:] = [[get_data(i, j) for j in columns] for i in rows]
    for i, j in factors: data[i, :] *= j
        
    if tickmarks is None: 
        tickmarks = [
            bst.plots.rounded_tickmarks_from_data(
                i, step_min=step_min, N_ticks=8, lb_max=0, center=0,
                f=roundsigfigs, expand=expand,
                f_min=lambda x: np.percentile(x, 5),
                f_max=lambda x: np.percentile(x, 95),
            ) 
            for i in data
        ]

    x0 = len(configuration_names) - 0.5
    xf = len(columns) - 0.5
    for i in range(N_rows):
        ax = axes[i]
        plt.sca(ax)
        if combined:
            bst.plots.plot_vertical_line(x0)
            ax.axvspan(x0, xf, color=colors.purple_tint.tint(60).RGBn)
        plt.xlim(-0.5, xf)

    for j in range(N_cols):
        color_wheel.restart()
        for i in range(N_rows):
            ax = axes[i]
            plt.sca(ax)
            plot(data[i, j], j)
            plt.ylabel(ylabels[i])
    
    for i in range(N_rows):
        ax = axes[i]
        plt.sca(ax)
        yticks = tickmarks[i]
        plt.ylim([yticks[0], yticks[1]])
        if yticks[0] < 0.:
            bst.plots.plot_horizontal_line(0, color=CABBI_colors.black.RGBn, lw=0.8, linestyle='--')
        try:
            xticklabels = xtext if (ax in axes_box[-1] or i == N_rows - 1) else []
        except:
            xticklabels = xtext if i == N_rows - 1 else []
        bst.plots.style_axis(ax,  
            xticks = xticks,
            yticks = yticks,
            xticklabels= xticklabels, 
            ytick0=False,
            ytickf=False,
            offset_xticks=True,
            xrot=xrot,
        )
    for i in range(N_rows, nrows * ncols):
        ax = axes[i]
        plt.sca(ax)
        plt.axis('off')
    if fig is None:
        fig = plt.gcf()
    else:
        plt.subplots_adjust(hspace=0)
    fig.align_ylabels(axes)
    return fig, axes

def plot_competitive_biomass_yield_across_oil_content(
        fs=None,
    ):
    if fs is None: fs = 10
    set_figure_size(aspect_ratio=0.8, width=5.5)
    fig, axes = plt.subplots(ncols=1, nrows=2)
    set_font(size=fs)
    bioethanol_ax, microbial_oil_ax = axes.flatten()
    plt.sca(bioethanol_ax)
    _plot_competitive_biomass_yield_across_oil_content('O2')
    bst.plots.style_axis(
        xticklabels=False,
        ytick0=False,
        ytickf=True,
    )
    plt.text(
        8.5, 43.5, '(A) Bioethanol', weight='bold',
        ha='center', va='center', fontsize=12,
        c=colors.neutral.shade(50).RGBn,
    )
    _add_lines_biomass_yield_vs_oil_content()
    plt.sca(microbial_oil_ax)
    _plot_competitive_biomass_yield_across_oil_content('O8')
    plt.xlabel('Oil content [dry wt. %]')
    bst.plots.style_axis(
        ytick0=False,
        ytickf=True,
    )
    plt.text(
        8.5, 43.5, '(B) Microbial oil', weight='bold', 
        ha='center', va='center', fontsize=12,
        c=colors.neutral.shade(50).RGBn,
    )
    _add_lines_biomass_yield_vs_oil_content()
    plt.subplots_adjust(hspace=0.05, left=0.1, right=0.96, bottom=0.10, top=0.95)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'competitive_biomass_yield_MCAC.{i}')
        plt.savefig(file, transparent=True)

def _add_lines_biomass_yield_vs_oil_content():
    df = cane.get_composition_data()
    txtbox = dict(boxstyle='round', facecolor=colors.neutral.shade(20).RGBn, 
                  edgecolor='None', alpha=0.9, pad=0.2)
    for name, color in zip(df.index, line_color_wheel):
        line = df.loc[name]
        oil = line['Stem oil (dw)']['Mean'] * 100
        biomass = line['Biomass yield (dry MT/ha)']['Mean']
        # breakpoint()
        bst.plots.plot_scatter_points(
            [oil], [biomass], marker='X', s=100, color=color.RGBn,
            edgecolor=edgecolor, clip_on=False, zorder=3
        )
        if name == '316':
            oil += 0.3
        if name == '19B':
            oil -= 0.3
            biomass += 0.3
        plt.text(
            oil + 0.2, biomass + 2, name, weight='bold', c=color.RGBn,
            bbox=txtbox,
        )

def _plot_competitive_biomass_yield_across_oil_content(
        configuration=None,
    ):
    if configuration is None: configuration = 'O2'
    file = monte_carlo_file(configuration, across_lines=False, across_oil_content='oilcane vs sugarcane')
    df = pd.read_excel(file, sheet_name=features.competitive_biomass_yield.short_description, index_col=0)
    df = df.dropna()
    oil_content = np.array(df.columns) * 100
    plt.ylabel(f"Biomass yield [dry-{format_units('MT/ha')}]")
    if configuration == 'O2':
        oil_content = oil_content[1:]
        biomass_yield = df.iloc[:, 1:]
        biomass_yield_p50 = bst.plots.plot_montecarlo_across_coordinate(oil_content, biomass_yield)[2]
        bst.plots.plot_vertical_line(oil_content[0])
        coeff = np.polyfit(oil_content, biomass_yield_p50, 3)
        # for i, j in zip(oil_content, biomass_yield_p50):
        #     print(i, j, np.polyval(coeff, i))
        target = cane.Biorefinery.baseline_dry_biomass_yield
        try:
            competitive_oil_content = flx.aitken_secant(lambda x: np.polyval(coeff, x) - target, x0=4)
        except:
            breakpoint()
        # print(competitive_oil_content, np.polyval(coeff, competitive_oil_content), target)
        bst.plots.plot_vertical_line(competitive_oil_content)
        plt.fill_between([oil_content[0], competitive_oil_content], 0, 50,
                          color=(*colors.CABBI_grey.tint(60).RGBn, 0.9),
                          zorder=0)
        plt.text(
            (oil_content[0] + competitive_oil_content) / 2, 5, 'Configuration\n trade-offs', c=colors.neutral.shade(50).RGBn,
            ha='center', va='center', fontsize=12,
        )
        plt.text(
            0.25, 12, 'Infeasible', c=colors.neutral.shade(50).RGBn,
            ha='left', va='center', fontsize=12, rotation='vertical',
        )
    else:
        biomass_yield_p50 = bst.plots.plot_montecarlo_across_coordinate(oil_content, df)[2]
    plt.xlim([0, 10])
    plt.ylim([0, 50])
    bst.plots.annotate_line(
        'Financially\ncompetitive\ntargets', 6, oil_content, biomass_yield_p50,
        dy=6, dy_text=0.8, position='over'
    )

def plot_competitive_microbial_oil_yield_across_oil_content(
        configuration=None,
    ):
    if configuration is None: configuration = 'O7'
    file = monte_carlo_file(configuration, across_lines=False, across_oil_content='microbial oil vs bioethanol')
    df = pd.read_excel(file, sheet_name=features.competitive_microbial_oil_yield.short_description, index_col=0)
    df = df.dropna(axis=0)
    fig = plt.figure()
    oil_fraction = np.array(df.columns) * 100
    plt.ylabel('Competitive microbial oil yield [wt %]')
    bst.plots.plot_montecarlo_across_coordinate(oil_fraction, df)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'competitive_microbial_oil_yield_MCAC_{configuration}.{i}')
        plt.savefig(file, transparent=True)

def plot_lines_monte_carlo(
        configurations=None, metrics=None, labels=None, tickmarks=None, 
        ncols=1, expand=None, step_min=None, xrot=None, color_wheel=None,
    ):
    if configurations is None: configurations = ['O2', 'O8']
    df = cane.get_composition_data()
    columns = df.index
    rows, ylabels = zip(*[mc_line_metric_settings[i] for i in metrics])
    if color_wheel is None: color_wheel = line_color_wheel
    nrows = len(rows)
    ncols = len(configurations)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows)
    plt.subplots_adjust(wspace=0.45)
    nmc = len(columns)    
    xtext = labels or columns
    N_marks = len(xtext)
    xticks = tuple(range(N_marks))
    
    def get_data(metric, line, configuration):
        df = get_line_monte_carlo(line, configuration, metric)
        values = df.values
        return values
    
    def plot(arr, position):
        if arr.ndim == 2:
            N = arr.shape[1]
            width = 0.618 / N
            boxwidth = 0.618 / (N + 1/N)
            plots = []
            for i in range(N):
                color = color_wheel.next()
                boxplot = monte_carlo_box_plot(
                    data=arr[:, i], positions=[position + (i-(N-1)/2)*width], 
                    light_color=color.RGBn, 
                    dark_color=color.shade(60).RGBn,
                    width=boxwidth,
                    hatch=getattr(color, 'hatch', None),
                )
                plots.append(boxplot)
            return plots
        else:
            color = color_wheel.next()
            return monte_carlo_box_plot(
                data=arr, positions=[position], 
                light_color=color.RGBn, 
                dark_color=color.shade(60).RGBn,
                width=0.618,
            )
    
    data = np.zeros([nrows, nmc], dtype=object)
    data[:] = [[np.array([get_data(i, j, k) for k in configurations]) for j in columns] for i in rows]
    
    if tickmarks is None: 
        tickmarks = [
            bst.plots.rounded_tickmarks_from_data(
                i, step_min=step_min, N_ticks=8, lb_max=0, center=0,
                f=roundsigfigs, expand=expand,
                f_min=lambda x: np.percentile(x, 2),
                f_max=lambda x: np.percentile(x, 98),
            ) 
            for i in data
        ]

    xf = len(columns) - 0.5
    for i in range(nrows):
        for k in range(ncols):
            color_wheel.restart()
            for j in range(nmc):
                ax = axes[i, k]
                plt.sca(ax)
                plt.xlim(-0.5, xf)
                plot(data[i, j][k], j)
                if k == 0: plt.ylabel(ylabels[i])
    
    titles = ['Bioethanol', 'Microbial oil']
    for i in range(nrows):
        for j in range(ncols):
            ax = axes[i, j]
            plt.sca(ax)
            if i == 0: ax.set_title(titles[j])
            yticks = tickmarks[i]
            plt.ylim([yticks[0], yticks[1]])
            if yticks[0] < 0.:
                bst.plots.plot_horizontal_line(0, color=CABBI_colors.black.RGBn, lw=0.8, linestyle='--')
            xticklabels = xtext if i == nrows - 1 else []
            yticklabels = j == 0
            bst.plots.style_axis(ax,  
                xticks = xticks,
                yticks = yticks,
                xticklabels= xticklabels, 
                yticklabels=yticklabels,
                ytick0=False,
                ytickf=False,
                offset_xticks=False,
                xrot=xrot,
            )
    if fig is None:
        fig = plt.gcf()
    else:
        plt.subplots_adjust(hspace=0)
        plt.subplots_adjust(wspace=0)
    fig.align_ylabels(axes)
    return fig, axes

#%% Spearman

def plot_spearman(configurations, labels=None, metric=None, 
                  kind=None, with_units=None, legend=None, legend_kwargs=None, **kwargs):
    if kind is None: kind = 'TEA'
    if with_units is None: with_units = True
    if legend is None: legend = True
    if metric is None:
        if kind == 'TEA':
            metric = MFPP
            metric_name = metric.name
        elif kind == 'LCA':
            metric = GWP_economic
            metric_name = r'GWP$_{\mathrm{economic}}$'
        else:
            raise ValueError(f"invalid kind '{kind}'")
    else:
        if metric == 'MFPP':
            metric = MFPP
        elif metric == 'GWP':
            metric = GWP_economic
        metric_name = metric.name
    stream_price = format_units('USD/L')
    USD_MT = format_units('USD/MT')
    ng_price = format_units('USD/m^3')
    electricity_price = format_units('USD/kWhr')
    operating_days = format_units('day/yr')
    capacity = format_units('10^6 MT/yr')
    titer = format_units('g/L')
    productivity = format_units('g/L/hr')
    material_GWP = '$\\mathrm{kg} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{kg}^{-1}$'
    feedstock_GWP = '$\\mathrm{g} \\cdot \\mathrm{CO}_{2}\\mathrm{eq} \\cdot \\mathrm{kg}^{-1}$'
    index, ignored_list = zip(*[
         ('Crushing mill oil recovery [60 $-$ 90 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Saccharification oil recovery [70 $-$ 90 %]', ['S2', 'S1', 'S2*', 'S1*', 'O1', 'O1*']),
        (f'Cane operating days [120 $-$ 180 {operating_days}]', []),
        (f'Sorghum operating days [30 $-$ 60 {operating_days}]', ['S2', 'S1', 'O1', 'O2']),
        (f'Crushing capacity [1.2 $-$ 2.0 {capacity}]', []),
        (f'Cellulosic ethanol price [0.655, 1.02, 1.03 {stream_price}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Advanced ethanol price [0.414, 0.599, 0.705 {stream_price}]', []),
        (f'Biodiesel price [0.800, 0.801, 1.57 {stream_price}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Natural gas price [0.105, 0.122, 0.175 {ng_price}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Electricity price [0.0583, 0.065, 0.069 {electricity_price}]', ['S2', 'O2', 'S2*', 'O2*']),
         ('IRR [10 $-$ 15 %]', []),
        (f'Crude glycerol price [100 $-$ 220 {USD_MT}]', ['S2', 'S1', 'S2*', 'S1*']),
        (f'Pure glycerol price [488 $-$ 812 {USD_MT}]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Saccharification reaction time [54 $-$ 90 hr]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Cellulase price [159 $-$ 265 {USD_MT}]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cellulase loading [1.5 $-$ 2.5 wt. % cellulose]', ['S1', 'O1', 'S1*', 'O1*']),
         ('PTRS base cost [14.9 $-$ 24.7 MMUSD]', ['S1', 'O1', 'S1*', 'O1*']),
         # ('Pretreatment reactor system base cost [14.9 $-$ 24.7 MMUSD]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane glucose yield [85 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Sorghum glucose yield [85 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane xylose yield [65 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Sorghum xylose yield [65 $-$ 97.5 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Glucose to ethanol yield [90 $-$ 95 %]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Xylose to ethanol yield [50 $-$ 95 %]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Titer [65 $-$ 130 {titer}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Productivity [1.0 $-$ 2.0 {productivity}]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Cane PL content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Sorghum PL content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Cane FFA content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Sorghum FFA content [7.5 $-$ 12.5 %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Cane oil content [5 $-$ 15 dry wt. %]', ['S2', 'S1', 'S2*', 'S1*']),
         ('Relative sorghum oil content [-3 $-$ 0 dry wt. %]', ['S2', 'S1', 'S2*', 'S1*', 'O2', 'O1']),
         ('TAG to FFA conversion [17.25 $-$ 28.75 % theoretical]', ['S1', 'O1', 'S1*', 'O1*']),
        # TODO: change lower upper values to baseline +- 10%
        (f'Feedstock GWPCF [26.3 $-$ 44.0 {feedstock_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Methanol GWPCF [0.338 $-$ 0.563 {material_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Pure glycerine GWPCF [1.25 $-$ 2.08 {material_GWP}]', ['S1', 'S2', 'S1*', 'S2*']),
        (f'Cellulase GWPCF [6.05 $-$ 10.1 {material_GWP}]', ['S1', 'O1', 'S1*', 'O1*']),
        (f'Natural gas GWPCF [0.297 $-$ 0.363 {material_GWP}]', ['S1', 'O1', 'S1*', 'O1*']),
         ('Corporate income tax [21 $-$ 28 %]', []),
    ])
    if not with_units: index = [i.split(' [')[0] for i in index]
    ignored_dct = {
        'S1': [],
        'O1': [],
        'S2': [],
        'O2': [],
        'S1*': [],
        'O1*': [],
        'S2*': [],
        'O2*': [],
    }
    for i, ignored in enumerate(ignored_list):
        for name in ignored: ignored_dct[name].append(i)
        index_name = index[i]
        if kind == 'LCA':
            for term in ('cost', 'price', 'IRR', 'time', 'capacity', 'operating'):
                if term in index_name:
                    for name in ignored_dct: ignored_dct[name].append(i)
                    break
        elif kind == 'TEA':
            if 'GWP' in index_name:
                for name in ignored_dct: ignored_dct[name].append(i)
        else:
            raise ValueError(f"invalid kind '{kind}'")
    
    rhos = []
    for name in configurations:
        file = spearman_file(name)
        try: 
            df = pd.read_excel(file, header=[0, 1], index_col=[0, 1])
        except: 
            warning = RuntimeWarning(f"file '{file}' not found")
            warn(warning)
            continue
        s = df[metric.index]
        s.iloc[ignored_dct[name]] = 0.
        rhos.append(s)
    color_wheel = [CABBI_colors.orange, CABBI_colors.green_soft, CABBI_colors.blue, CABBI_colors.brown]
    fig, ax = bst.plots.plot_spearman_2d(rhos, index=index,
                                         color_wheel=color_wheel,
                                         name=metric_name,
                                         **kwargs)
    if legend:
        if legend_kwargs is None:
            legend_kwargs = {'loc': 'lower left'}
        plt.legend(
            handles=[
                mpatches.Patch(
                    color=color_wheel[i].RGBn, 
                    label=labels[i] if labels else format_name(configurations[i])
                )
                for i in range(len(configurations))
            ], 
            **legend_kwargs,
        )
    return fig, ax

# %% Other

def plot_configuration_breakdown(name, across_coordinate=False, **kwargs):
    oc = cane.Biorefinery(name)
    if across_coordinate:
        return bst.plots.plot_unit_groups_across_coordinate(
            cane.set_cane_oil_content,
            [5, 7.5, 10, 12.5],
            'Feedstock oil content [dry wt. %]',
            oc.unit_groups,
            colors=[area_colors[i.name].RGBn for i in oc.unit_groups],
            hatches=[area_hatches[i.name] for i in oc.unit_groups],
            **kwargs,
        )
    else:
        def format_total(x):
            if x < 1e3:
                return format(x, '.3g')
            else:
                x = int(x)
                n = 10 ** (len(str(x)) - 3)
                value = int(round(x / n) * n)
                return format(value, ',')
        for i in oc.unit_groups: 
            if i.name == 'EtOH prod.':
                i.name = 'Ethanol production'
            elif i.name == 'Oil ext.':
                i.name = 'Oil extraction'
            elif i.name == 'Biod. prod.':
                i.name = 'Biodiesel production'
            i.metrics[0].name = 'Inst. eq.\ncost'
            i.metrics[3].name = 'Elec.\ncons.'
            i.metrics[4].name = 'Mat.\ncost'
        
        return bst.plots.plot_unit_groups(
            oc.unit_groups,
            colors=[area_colors[i.name].RGBn for i in oc.unit_groups],
            hatches=[area_hatches[i.name] for i in oc.unit_groups],
            format_total=format_total,
            fraction=True,
            legend_kwargs=dict(
                loc='lower center',
                ncol=4,
                bbox_to_anchor=(0, -0.52),
                labelspacing=1.5, handlelength=2.8,
                handleheight=1, scale=0.8,
            ),
            **kwargs,
        )
    
# def plot_monte_carlo_across_oil_content(kind=0, derivative=False):
#     MFPP, TCI, *production, electricity_production, natural_gas_consumption = tea_monte_carlo_metric_mockups
#     rows = [MFPP, TCI, production]
#     if kind == 0:
#         columns = across_oil_content_names
#     elif kind == 1:
#         columns = across_oil_content_agile_names
#     elif kind == 2:
#         columns = across_oil_content_comparison_names
#     elif kind == 3:
#         columns = across_oil_content_agile_comparison_names
#     elif kind == 4:
#         columns = across_oil_content_agile_direct_comparison_names
#     else:
#         raise NotImplementedError(str(kind))
#     if derivative:
#         x = 100 * (oil_content[:-1] + np.diff(oil_content) / 2.)
#         ylabels = [
#             f"MFPP der. [{format_units('USD/MT')}]",
#             f"TCI der. [{format_units('10^6*USD')}]",
#             f"Production der. [{format_units('L/MT')}]"
#         ]
#     else:
#         x = 100 * oil_content
#         ylabels = [
#             f"MFPP$\backprime$ [{format_units('USD/MT')}]",
#             f"TCI [{format_units('10^6*USD')}]",
#             f"Production [{format_units('L/MT')}]"
#         ]
#     N_cols = len(columns)
#     N_rows = len(rows)
#     fig, axes = plt.subplots(ncols=N_cols, nrows=N_rows)
#     data = np.zeros([N_rows, N_cols], dtype=object)
    
#     def get_data(metric, name):
#         if isinstance(metric, bst.Variable):
#             return get_monte_carlo_across_oil_content(name, metric, derivative)
#         else:
#             return [get_data(i, name) for i in metric]
    
#     data = np.array([[get_data(i, j) for j in columns] for i in rows])
#     tickmarks = [None] * N_rows
#     get_max = lambda x: max([i.max() for i in x]) if isinstance(x, list) else x.max()
#     get_min = lambda x: min([i.min() for i in x]) if isinstance(x, list) else x.min()
#     N_ticks = 5
#     for r in range(N_rows):
#         lb = min(min([get_min(i) for i in data[r, :]]), 0)
#         ub = max([get_max(i) for i in data[r, :]])
#         diff = 0.1 * (ub - lb)
#         ub += diff
#         if derivative:
#             lb = floor(lb)
#             ub = ceil(ub)
#             step = (ub - lb) / (N_ticks - 1)
#             tickmarks[r] = [0, 1] if step == 0 else [int(lb + step * i) for i in range(N_ticks)]
#         else:
#             if rows[r] is MFPP:
#                 if kind == 0 or kind == 1:
#                     tickmarks[r] = [-20, 0, 20, 40, 60]
#                 elif kind == 2:
#                     tickmarks[r] = [-20, -10, 0, 10, 20]
#                 elif kind == 3:
#                     tickmarks[r] = [-10, 0, 10, 20, 30]
#                 elif kind == 4:
#                     tickmarks[r] = [-5, 0, 5, 10, 15]
#                 continue
#             lb = floor(lb / 15) * 15
#             ub = ceil(ub / 15) * 15
#             step = (ub - lb) / (N_ticks - 1)
#             tickmarks[r] = [0, 1] if step == 0 else [int(lb + step * i) for i in range(N_ticks)]
#     color_wheel = CABBI_colors.wheel()
#     for j in range(N_cols):
#         color_wheel.restart()
#         for i in range(N_rows):
#             arr = data[i, j]
#             ax = axes[i, j]
#             plt.sca(ax)
#             percentiles = plot_monte_carlo_across_coordinate(x, arr, color_wheel)
#             if i == 0: ax.set_title(format_name(columns[j]))
#             xticklabels = i == N_rows - 1
#             yticklabels = j == 0
#             if xticklabels: plt.xlabel('Oil content [dry wt. %]')
#             if yticklabels: plt.ylabel(ylabels[i])
#             bst.plots.style_axis(ax,  
#                                  xticks = [5, 10, 15],
#                                  yticks = tickmarks[i],
#                                  xticklabels= xticklabels, 
#                                  yticklabels= yticklabels,
#                                  ytick0=False)
#     for i in range(N_cols): fig.align_ylabels(axes[:, i])
#     plt.subplots_adjust(hspace=0.1, wspace=0.1)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.biodiesel import (
    create_lipid_pretreatment_system as create_oil_pretreatment_system,
    create_transesterification_and_biodiesel_separation_system,
)
from .juicing import (
    create_feedstock_handling_system,
    create_juicing_system,
)
from .fermentation import create_sucrose_fermentation_system
from .lipid_extraction import create_lipid_exctraction_system
from ..data import microbial_oil_baseline as perf
from .. import units
from .. import streams as s

__all__ = (
    'create_oilcane_to_biodiesel_1g',
)

@SystemFactory(
    ID='oilcane_sys',
    ins=[s.oilcane],
    outs=[s.biodiesel, s.crude_glycerol, s.vinasse],
)
def create_oilcane_to_biodiesel_1g(
            ins, outs, fed_batch=True, WWT_kwargs=None,
        ):
    oilcane, = ins
    biodiesel, crude_glycerol, vinasse = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        ins=oilcane,
        outs='',
        mockup=True,
        area=100
    )
    
    ### Oil and juice separation ###
    
    juicing_sys, jdct = create_juicing_system(
        ins=feedstock_handling_sys-0,
        outs=['', '', 'fiber_fines'],
        pellet_bagasse=False,
        dry_bagasse=True,
        mockup=True,
        udct=True,
        area=200,
    )
    screened_juice, bagasse, fiber_fines = juicing_sys.outs
    # bagasse_pelleting_sys = create_bagasse_pelleting_system(None, bagasse, area=200, mockup=True)
    # pelleted_bagasse, = bagasse_pelleting_sys.outs
    jdct['S201'].isplit['Lipid'] = 1. # Vibrating screen
    crushing_mill = jdct['U201']
    crushing_mill.isplit['Lipid'] = 0.90
    
    ### Ethanol section ###
    if fed_batch:
        biomass_coeff = perf.fed_batch_biomass_growth_coefficient_mean
        lipid_yield = perf.fed_batch_lipid_yield_mean
        titer = perf.fed_batch_titer_mean
        productivity = perf.fed_batch_productivity_mean
    else:
        biomass_coeff = perf.batch_biomass_growth_coefficient_mean
        lipid_yield = perf.batch_lipid_yield_mean
        titer = perf.batch_titer_mean
        productivity = perf.batch_productivity_mean
    
    fermrxn = tmo.Rxn('O2 + Glucose -> H2O + TAG', 'Glucose', 1., correct_atomic_balance=True)
    fermrxn.product_yield('TAG', basis='wt', product_yield=lipid_yield)
    cellmass_rxn = tmo.Rxn(
        'Glucose -> H2O + CO2 + Yeast', 'Glucose', biomass_coeff, 
        correct_atomic_balance=True
    )
    cellmass_rxn.product_yield('Yeast', 'wt', biomass_coeff)
    combustion = tmo.Rxn('Glucose + O2 -> CO2 + H2O', 'Glucose', 1. - cellmass_rxn.X,
                         correct_atomic_balance=True)
    growrxn = cellmass_rxn + combustion
    growrxn.X = 0.999
    fermentation_sys, epdct = create_sucrose_fermentation_system(
        ins=[screened_juice],
        scrubber=False,
        SeedTrain=units.SeedTrain,
        Fermentor=units.AeratedFermentation,
        seed_train_reaction=growrxn,
        fermentation_reaction=fermrxn,
        cell_growth_reaction=growrxn,
        fed_batch=fed_batch,
        titer=titer,
        productivity=productivity,
        product_group='Lipid',
        fermentation_kwargs={},
        mockup=True,
        area=300,
        add_urea=False,
        udct=True,
    )
    product, condensate, vent = fermentation_sys.outs
    post_fermentation_oil_separation_sys = create_lipid_exctraction_system(
        ins=product,
        mockup=True,
        area=300,
    )
    oil, cellmass, thick_vinasse = post_fermentation_oil_separation_sys.outs
    oil_pretreatment_sys, oil_pretreatment_dct = create_oil_pretreatment_system(
        ins=oil,
        mockup=True,
        outs=['', 'polar_lipids', ''],
        area=600,
        udct=True,
    )
    bst.Mixer(300, [thick_vinasse, condensate], vinasse)
    oil, polar_lipids, wastewater = oil_pretreatment_sys.outs
    
    # Fresh degummed oil
    transesterification_and_biodiesel_separation_sys, tbdct = create_transesterification_and_biodiesel_separation_system(
        ins=oil, 
        outs=[biodiesel, crude_glycerol, ''],
        mockup=True,
        area=600,
        udct=True,
    )
    bst.Mixer(600, [transesterification_and_biodiesel_separation_sys-2, wastewater], 'wastewater')

    ### Facilities ###
    if WWT_kwargs:
        WWT = True
        WWT_kwargs['area'] = 1000
    else:
        WWT = False
    
    u = f.unit
    bst.create_all_facilities(
        feedstock=None,
        HXN_kwargs=dict(
            ID=900,
            ignored=lambda: [u.E301, u.D601.boiler, u.D602.boiler, u.H601, u.H602, u.H603, u.H604, oil_pretreatment_dct['F3']],
            Qmin=1e5,
            acceptable_energy_balance_error=0.01,
        ),
        CHP_kwargs=dict(area=700),
        WWT_kwargs=WWT_kwargs,
        WWT=WWT,
        area=800,
    )



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
#                     Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst

__all__ = (
    'BiorefinerySettings',
    'default_feedstock_composition',
    'default_load_utility_settings',
    'default_stream_GWP_CFs',
    'default_stream_prices',
    )


class BiorefinerySettings:
    '''
    Containing the general parameters needed in biorefinery design/TEA/LCA
    that can be modified to update the biorefinery.
    '''
    
    def __init__(
            self,
            system_ID='corn_sys',
            CEPCI=567, # 2013 chemical engineering plant capital index
            load_utility_settings=None,
            feedstock_composition={},
            feedstock_hourly_mass_flow=46375, # kg / hr
            process_parameters={},
            stream_prices={},
            stream_GWP_CFs={},
            other_unit_parameters={},
            tea_parameters={},
            **kwargs,
            ):
        self.system_ID = system_ID
        self.CEPCI = CEPCI
        self.load_utility_settings = load_utility_settings or default_load_utility_settings
        self.feedstock_composition = feedstock_composition or default_feedstock_composition
        self.feedstock_hourly_mass_flow = feedstock_hourly_mass_flow
        self.process_parameters = process_parameters or default_process_parameters
        self.stream_prices = stream_prices or default_stream_prices
        self.stream_GWP_CFs = stream_GWP_CFs or default_stream_GWP_CFs
        self.other_unit_parameters = other_unit_parameters
        self.tea_parameters = tea_parameters
        for k, v in kwargs: setattr(self, k, v)
        
    def load_process_settings(self):
        '''Set capital index and load utility settings'''
        bst.CE = self.CEPCI
        self.load_utility_settings()
        

def default_load_utility_settings(
        power_price=0.07, # Value from Chinmay's report; 0.07, in original
        power_GWP_CFs=(1., 1.,), # consumption, production #!!! needs to be updated
        ):
    bst.process_tools.default_utilities()
    bst.PowerUtility.price = power_price
    bst.PowerUtility.characterization_factors['GWP'] = power_GWP_CFs
    HeatUtility = bst.HeatUtility
    lps = HeatUtility.get_agent('low_pressure_steam')
    Water = lps.chemicals.Water
    lps.T = 152 + 273.15
    lps.P = Water.Psat(lps.T)
    lps.heat_transfer_efficiency = 1.0
    lps.regeneration_price = default_stream_prices['Steam'] * Water.MW
    for agent in HeatUtility.heating_agents:
        agent.regeneration_price = default_stream_prices['Steam'] * Water.MW
    cw = HeatUtility.get_agent('cooling_water')
    cw.regeneration_price = 0.073e-3 * Water.MW
    cw.T = 25. + 273.15


# Feedstock weight composition
default_feedstock_composition = {
    'Starch': 0.612,
    'Water': 0.15,
    'Fiber': 0.1067,
    'SolubleProtein': 0.034,
    'InsolubleProtein': 0.0493,
    'Oil': 0.034,
    'Ash': 0.014
    }


# Process parameters
default_process_parameters = {
    'slurry_solids_content': 0.311,   # g Corn / g Total
    'slurry_ammonia_loading': 0.002,  # g Ammonia / g dry Corn
    'slurry_lime_loading': 0.00012,   # g Lime / g dry Corn
    'liquefaction_alpha_amylase_loading': 0.00082,    # g Enzyme / g dry Corn
    'saccharification_sulfuric_acid_loading': 0.001, # g H2SO4 / g dry Corn
    'saccharification_gluco_amylase_loading': 0.0011, # g Enzyme / g dry Mash
    'scrubber_wash_water_over_vent': 1.21, # g Water / g Vent
    'yeast_loading': 8e-5, # g Yeast (5 wt %) / g Mash
    }

# Characterization factors for 100-year global warming potential
#!!! Numbers are all fake now
default_stream_GWP_CFs = {
    'Ethanol': 1,
    'Corn': 1,
    'DDGS': 1,
    'Yeast': 1,
    'Enzyme': 1,
    'Denaturant': 1,
    'Sulfuric acid': 1,
    'Ammonia': 1,
    'Caustic': 1,
    'Lime': 1,
    'Steam': 1,
    'Crude oil': 1,
}


# Raw material price (USD/kg)
default_stream_prices = {
    'Ethanol': 0.48547915353569393, 
    'Corn': 0.13227735731092652, # Value from Chinmay's report; 0.08476585075177462 in original
    'DDGS': 0.12026, # Value from Chinmay's report; 0.09687821462905594, in original
    'Yeast': 1.86, # 1.86 in Chinmay's report; 0.907310526171629 in original
    'Enzyme': 2.25, # Value from Chinmay's report; 3.20739717428309 in original
    'Denaturant': 0.4355069727002459,
    'Sulfuric acid': 0.08971711759613593,
    'Ammonia': 0.4485966110937889,
    'Caustic': 0.14952852932689323,
    'Lime': 0.19937504680689405,
    'Steam': 12.86e-3, # Value from Chinmay's report; 17.08e-3 in original lipidcane report 
    'Crude oil': 0.56,
}





# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autofunction:: biorefineries.cane.chemicals.create_sugarcane_chemicals
.. autofunction:: biorefineries.cane.chemicals.create_oilcane_chemicals
.. autofunction:: biorefineries.cane.chemicals.create_cellulosic_oilcane_chemicals
.. autofunction:: biorefineries.cane.chemicals.create_acyl_olein
.. autofunction:: biorefineries.cane.chemicals.create_acetyl_diolein

"""
import thermosteam as tmo
from thermosteam import functional as fn
from chemicals import atoms_to_Hill
from thermosteam.utils import chemical_cache
from biorefineries import cellulosic
import biosteam as bst

__all__ = (
    'create_sugarcane_chemicals',
    'create_oilcane_chemicals',
    'create_cellulosic_oilcane_chemicals',
    'create_acyl_olein',
    'create_acetyl_diolein',
)

@chemical_cache
def create_sugarcane_chemicals(yeast_includes_nitrogen=None):
    if yeast_includes_nitrogen is None: yeast_includes_nitrogen = False
    (Water, Ethanol, Glucose, Sucrose, H3PO4, P4O10, CO2, Octane, O2, N2, CH4) = chemicals = tmo.Chemicals(
        ['Water', 'Ethanol', 
         tmo.Chemical('Glucose', phase='l'),
         tmo.Chemical('Sucrose', phase='l'), 
         tmo.Chemical('H3PO4', phase='s'),
         tmo.Chemical('P4O10', phase='s'),
         tmo.Chemical('CO2', phase='g'), 
         'Octane', 
         tmo.Chemical('O2', phase='g'), 
         tmo.Chemical('N2', phase='g'), 
         tmo.Chemical('CH4', phase='g')]
    )
    Glucose.N_solutes = 1
    Sucrose.N_solutes = 2
    
    def create_new_chemical(ID, phase='s', **constants):
        chemical = tmo.Chemical(ID, phase=phase, phase_ref=phase, search_db=False, **constants)
        chemicals.append(chemical)
        return chemical
    
    Ash = create_new_chemical('Ash', MW=1.)
    Cellulose = create_new_chemical('Cellulose',
                                    formula="C6H10O5", # Glucose monomer minus water
                                    Hf=-975708.8)
    Hemicellulose = create_new_chemical('Hemicellulose',
                                        formula="C5H8O4", # Xylose monomer minus water
                                        Hf=-761906.4)
    Flocculant = create_new_chemical('Flocculant',
                                     MW=1.)
    Lignin = create_new_chemical('Lignin',
                                 formula='C8H8O3', # Vainillin
                                 Hf=-452909.632)
    Solids = create_new_chemical('Solids', MW=1.)
    Yeast = create_new_chemical(
        'Yeast', 
        formula='CH1.61O0.56N0.16' if yeast_includes_nitrogen else 'CH1.61O0.56',
        rho=1540,
        Cp=Glucose.Cp(298.15),
        default=True,
    )
    Yeast.Hf = Glucose.Hf / Glucose.MW * Yeast.MW # Same as glucose to ignore heats related to growth
    CaO = create_new_chemical('CaO', formula='CaO')

    
    ### Fill missing properties ###
    
    # Insolubles occupy a significant volume
    insoluble_solids = (Ash, Cellulose, Hemicellulose,
                        Flocculant, Lignin, Solids, Yeast, P4O10)
    
    # Solubles don't occupy much volume
    soluble_solids = (CaO, H3PO4, Glucose, Sucrose) 
    
    for chemical in insoluble_solids:
        V = fn.rho_to_V(rho=1540, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    for chemical in soluble_solids:
        V = fn.rho_to_V(rho=1e5, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    # Add constant models for molar heat capacity of solids
    # Lignocellulosic heat capacities:
    # https://link.springer.com/article/10.1007/s10853-013-7815-6
    # https://www.sciencedirect.com/science/article/pii/0032386182901252
    Ash.Cn.add_model(0.09 * 4.184 * Ash.MW) 
    CaO.Cn.add_model(1.02388 * CaO.MW) 
    Cellulose.Cn.add_model(1.364 * Cellulose.MW) 
    Hemicellulose.Cn.add_model(1.364 * Hemicellulose.MW)
    Flocculant.Cn.add_model(4.184 * Flocculant.MW)
    Lignin.Cn.add_model(1.364 * Lignin.MW)
    Solids.Cn.add_model(1.100 * Solids.MW)
    
    for chemical in chemicals: chemical.default()
    
    chemicals.compile()
    chemicals.set_synonym('Water', 'H2O')
    chemicals.set_synonym('Yeast', 'DryYeast')
    
    return chemicals

@chemical_cache
def create_acyl_olein(N_acyl):
    # Assume properties are similar between oleic and palmitic chains
    Hf_oleic_acid = -764.80e3
    Hf_triolein = -1776e3
    Hf_water = -285.825e3
    Hf_glycerol = -668.6e3
    # TAG + 3H2O -> GLY + 3FFA
    # Hrxn = (GLY + 3FFA) - (TAG + 3H2O)
    Hf_tag_to_ffa = (Hf_glycerol + 3 * Hf_oleic_acid) - (Hf_triolein + 3 * Hf_water)
    Hf_minus_one_acyl = Hf_tag_to_ffa / 3. 
    if N_acyl == 0:
        ID_model = 'Palmitic acid'
        ID = 'OleicAcid'
        Hf = Hf_oleic_acid
    elif N_acyl == 1:
        ID_model = 'Monopalmitin'
        ID = 'MonoOlein'
        # TAG + 2H2O -> MAG + 2FFA
        # Hrxn = MAG + 2FFA - (TAG + 2H2O) # Assume to be Hf_tag_to_ffa * 2 / 3
        # MAG = Hrxn - 2FFA + TAG + 2H2O
        Hf = 2 * Hf_minus_one_acyl - 2 * Hf_oleic_acid + Hf_triolein + 2 * Hf_water
    elif N_acyl == 2:
        ID_model = 'Dipalmitin'
        ID = 'DiOlein'
        # TAG + H2O -> DAG + FFA
        # Hrxn = DAG + FFA - (TAG + H2O) # Assume to be Hf_tag_to_ffa / 3
        # DAG = Hrxn - FFA + TAG + H2O
        Hf = Hf_minus_one_acyl - Hf_oleic_acid + Hf_triolein + Hf_water
    elif N_acyl == 3:
        ID_model = 'Tripalmitin'
        ID = 'TriOlein'
        Hf = Hf_triolein
    model = tmo.Chemical(ID_model).at_state(phase='l', copy=True)
    atoms = model.atoms
    atoms['C'] += N_acyl * 2
    atoms['H'] += N_acyl * 2
    formula = atoms_to_Hill(atoms)
    chemical = tmo.Chemical(ID, search_db=False, phase='l', phase_ref='l',
                            formula=formula, Hf=Hf)
    chemical._Dortmund = group_counts = model.Dortmund.copy()
    group_counts.set_group_counts_by_name({'CH=CH': N_acyl}, reset=False)
    chemical.V.add_model(fn.rho_to_V(rho=900, MW=chemical.MW))
    chemical.Cn.add_model(model.Cp(298.15) * chemical.MW)
    chemical.copy_models_from(model, ['mu', 'sigma', 'kappa'])
    if ID == 'MonoOlein': chemical.mu.add_model(lambda T: 0.0001)
    return chemical

@chemical_cache
def create_acetyl_diolein():
    # Assume properties are similar between oleic and palmitic chains
    Hf_oleic_acid = -764.80e3
    Hf_triolein = -1776e3
    Hf_acetic_acid = -483.58e3
    ID_model = 'Palmitin'
    ID = 'AcetylDiOlein'
    # Ac + TAG -> acTAG + FFA
    # Hrxn = acTAG + FFA - (TAG + Ac)  # Assume Hrxn is negligible
    # acTAG = Hrxn - FFA + TAG + Ac
    Hf = Hf_triolein - Hf_oleic_acid + Hf_acetic_acid
    model = tmo.Chemical(ID_model).at_state(phase='l', copy=True)
    formula = atoms_to_Hill({'C': 41, 'H':75, 'O': 6})
    chemical = tmo.Chemical(ID, search_db=False, phase='l', phase_ref='l',
                            formula=formula, Hf=Hf)
    chemical._Dortmund = group_counts = model.Dortmund.copy()
    group_counts.set_group_counts_by_name({'CH=CH': 2, 'CH':1, 'CH2': 28, 'CH3': 3, 'CH2COO': 2, 'COOH': 1}, reset=True)
    chemical.V.add_model(fn.rho_to_V(rho=900, MW=chemical.MW))
    chemical.mu.add_method(900 * 20.3e-6)
    chemical.Cn.add_method(model.Cp(298.15) * chemical.MW)
    chemical.copy_models_from(model, ['sigma', 'kappa'])
    return chemical

@chemical_cache
def create_oilcane_chemicals(yeast_includes_nitrogen=None):
    chemicals = create_sugarcane_chemicals(yeast_includes_nitrogen).copy()
    (Water, Ethanol, Glucose, Sucrose, H3PO4, P4O10, CO2, Octane, O2, N2, CH4, 
     Ash, Cellulose, Hemicellulose, Flocculant, Lignin, Solids, DryYeast, CaO) = chemicals
    
    ### Define common chemicals ###
    Biodiesel = tmo.Chemical('Biodiesel',
                             search_ID='Methyl oleate')
    Methanol = tmo.Chemical('Methanol')
    Glycerol = tmo.Chemical('Glycerol')
    chemicals.extend([Biodiesel, Methanol, Glycerol])
    
    ### Define new chemicals ###
    
    def create_new_chemical(ID, phase='s', **constants):
        solid = tmo.Chemical.blank(ID, phase=phase, **constants)
        chemicals.append(solid)
        return solid
    
    HCl = create_new_chemical('HCl', formula='HCl')
    NaOH = create_new_chemical('NaOH', formula='NaOH')
    NaOCH3 = create_new_chemical('NaOCH3', formula='NaOCH3')
    ### Fill missing properties ###
    
    # Solubles don't occupy much volume
    for chemical in (HCl, NaOH):
        V = fn.rho_to_V(rho=1e5, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    # Assume sodium methoxide has some of the same properties as methanol
    LiquidMethanol = Methanol.at_state(phase='l', copy=True)
    NaOCH3.copy_models_from(LiquidMethanol, ['V', 'sigma', 'kappa', 'Cn'])
    chemicals.extend([
        tmo.Chemical('Phosphatidylinositol', formula='C47H83O13P',
                     search_db=False, CAS='383907-36-6', default=True,
                     Hf=-1.779e6, # Assume same as TAG on a weight basis
                     aliases={'PL', 'PolarLipid'}, phase='l'),
        # tmo.Chemical('BetaSitosterol', search_ID='83-46-5',
        #               phase='l', Hf=-1000. * (533.79 + 93.90),
        #               aliases=['Sterol']),
        create_acyl_olein(0),
        create_acyl_olein(1),
        create_acyl_olein(2),
        create_acyl_olein(3),
        tmo.Chemical('Acetone')
    ])
    for chemical in chemicals: chemical.default()
    chemicals.compile()
    chemicals.set_synonym('OleicAcid', 'FFA')
    chemicals.set_synonym('MonoOlein', 'MAG')
    chemicals.set_synonym('DiOlein', 'DAG')
    chemicals.set_synonym('TriOlein', 'TAG')
    chemicals.define_group('Lipid', ('PL', 'FFA', 'MAG', 'DAG', 'TAG'))
    chemicals.define_group('Oil', ('PL', 'FFA', 'MAG', 'DAG', 'TAG'))
    chemicals.set_synonym('Water', 'H2O')
    chemicals.set_synonym('Yeast', 'DryYeast')
    return chemicals

@chemical_cache
def create_cellulosic_oilcane_chemicals(yeast_includes_nitrogen=None):
    oilcane_chemicals = create_oilcane_chemicals(yeast_includes_nitrogen)
    cellulosic_chemicals = cellulosic.create_cellulosic_ethanol_chemicals()
    removed = {'SuccinicAcid', 'H2SO4', 'Z_mobilis', 'Oil'}
    chemicals = tmo.Chemicals([
        i for i in (oilcane_chemicals.tuple + cellulosic_chemicals.tuple) if i.ID not in removed
    ])
    chemicals.extend([
        create_acetyl_diolein(),
        tmo.Chemical('Urea', default=True, phase='l'),
    ])
    chemicals.extend(
        bst.wastewater.high_rate.create_missing_wwt_chemicals(chemicals)
    )
    chemicals.compile()
    chemicals.set_synonym('AcetylDiOlein', 'AcTAG')
    chemicals.define_group('Lipid', ['PL', 'FFA', 'MAG', 'DAG', 'TAG', 'AcTAG'])
    chemicals.define_group('lipid', ['PL', 'FFA', 'MAG', 'DAG', 'TAG', 'AcTAG'])
    chemicals.define_group('Oil', ['PL', 'FFA', 'MAG', 'DAG', 'TAG'])
    chemicals.set_synonym('Yeast', 'DryYeast')
    chemicals.set_synonym('Yeast', 'Cellmass')
    chemicals.AcetylDiOlein.V.method_P = chemicals.TriOlein.V.method_P = None
    chemicals.set_synonym('Cellmass', 'Cells')
    chemicals.set_synonym('Cellmass', 'cellmass')
    chemicals.set_synonym('TriOlein', 'TAG')
    return chemicals


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""

import biosteam as bst
from thermosteam import Stream
from biorefineries.cellulosic import units
from biorefineries.cellulosic import streams as s


__all__ = (
    'create_cofermentation_system',
)

@bst.SystemFactory(
    ID='cofermentation_sys',
    ins=[s.slurry, s.DAP, s.CSL],
    outs=[s.vent, s.beer, s.lignin],
)
def create_cofermentation_system(
        ins, outs, CoFermentation=None, SeedTrain=None,
        include_scrubber=None,
        seed_train_reactions=None,
        cofermentation_reactions=None,
        add_nutrients=True,
    ):
    """
    Create co-fermentation system that includes a pressure filter to separate
    lignin before fermentation.
    
    """
    slurry, DAP, CSL = ins
    vent, beer, lignin = outs
    if not SeedTrain: SeedTrain = units.SeedTrain
    if not CoFermentation: CoFermentation = units.CoFermentation
    has_vent = CoFermentation._N_outs == 2
    if has_vent:
        if include_scrubber is None: include_scrubber = True
    else:
        include_scrubber = False
        outs.remove(vent)
        
    if add_nutrients:
        DAP1 = Stream('DAP1',
                        DAP=26,
                        units='kg/hr')
        DAP2 = Stream('DAP2',
                        DAP=116,
                        units='kg/hr')
        CSL1 = Stream('CSL1',
                        CSL=211,
                        units='kg/hr')
        CSL2 = Stream('CSL2',
                        CSL=948,
                        units='kg/hr')
        DAP_storage = units.DAPStorageTank('DAP_storage', DAP)
        S301 = bst.MockSplitter('S301', DAP_storage-0, outs=(DAP1, DAP2))
        CSL_storage = units.CSLStorageTank('CSL_storage', CSL)
        S302 = bst.MockSplitter('S302', CSL_storage-0, outs=(CSL1, CSL2))
        nutrients_1 = (CSL1, DAP1)
        nutrients_2 = (CSL2, DAP2)
    else:
        ins.remove(DAP)
        ins.remove(CSL)
        nutrients_1 = nutrients_2 = ()
    
    S303 = bst.PressureFilter('S303', slurry, (lignin, ''))
    S304 = bst.Splitter('S304', S303-1, split=0.1)
    R302 = SeedTrain('R302', (S304-0, *nutrients_1), reactions=seed_train_reactions)
    
    if add_nutrients:
        @R302.add_specification(run=True)
        def adjust_CSL_and_DAP_feed_to_seed_train():
            feed, CSL1, DAP1 = R302.ins
            CSL1.imass['CSL'] = 0.0050 * feed.F_mass
            DAP1.imass['DAP'] = 0.67 * feed.F_vol
    
    T301 = units.SeedHoldTank('T301', R302-1)
    R303 = CoFermentation('R303', (S304-1, T301-0, *nutrients_2),
                          outs=('', ''), cofermentation=cofermentation_reactions)
    
    if add_nutrients:
        @R303.add_specification(run=True)
        def adjust_CSL_and_DAP_feed_to_fermentation():
            feed, seed, CSL2, DAP2, *other = R303.ins
            CSL2.imass['CSL'] = 0.0025 * feed.F_mass
            DAP2.imass['DAP'] = 0.33 * feed.F_vol
            S301.ins[0].mix_from(S301.outs)
            S302.ins[0].mix_from(S302.outs)
            DAP_storage.ins[0].copy_like(DAP_storage.outs[0])
            CSL_storage.ins[0].copy_like(CSL_storage.outs[0])
        
    T302 = units.BeerTank('T302', outs=beer)
    
    if include_scrubber:
        stripping_water = Stream('stripping_water',
                                 Water=26836,
                                 units='kg/hr')
        M401 = bst.Mixer('M401', (R303-1, None))
        M304 = bst.Mixer('M304', (R302-0, R303-0))
        D401 = bst.VentScrubber('D401', (stripping_water, M304-0), (vent, ''),
                                gas=('CO2', 'NH3', 'O2'))
        D401-1-1-M401-0-T302
    
        stripping_water_over_vent = stripping_water.imol['Water'] / 21202.490455845436
        @D401.add_specification(run=True)
        def update_stripping_water():
            stripping_water, vent = D401.ins
            stripping_water.imol['Water'] = stripping_water_over_vent * vent.F_mass

    elif has_vent:
        M304 = bst.Mixer('M304', (R302-0, R303-0), vent)
        R303-1-T302
    else:
        R303-0-T302



# -*- coding: utf-8 -*-
"""s
"""

import biosteam as bst
from thermosteam import Stream
import thermosteam.reaction as rxn
from biorefineries.cellulosic import streams as s

__all__ = (
    'create_ammonia_fiber_expansion_pretreatment_system',
)

@bst.SystemFactory(
    ID='AFEX_pretreatment_sys',
    ins=[s.switchgrass, s.ammonia], # Price of switchgrass, table 4, Madhu Khanna et al. (Costs of producing miscanthus and switchgrass for bioenergy in Illinois);  https://www.sciencedirect.com/science/article/pii/S096195340700205X?casa_token=KfYfzJtDwv0AAAAA:OqeJmpofk1kIgFk2DcUvXNG35qYwlWvPKZ7ENI3R6RUKeoahiTDpOhhd_mpLtRthTGuXJKDzMOc
    outs=[dict(ID='pretreated_biomass'),],
)
def create_ammonia_fiber_expansion_pretreatment_system(
        ins, outs,
        solids_loading=0.20,
        ammonia_loading=2, # g ammonia / g dry feedstock
        T_pretreatment_reactor=273.15 + 100.,
        residence_time=0.5,
        pretreatment_reactions=None,
        neutralize=True,
    ):
    feedstock, ammonia = ins
    pretreated_biomass, = outs
    warm_process_water = Stream('warm_process_water',
                              T=368.15,
                              P=4.7*101325,
                              Water=1)
    pretreatment_steam = Stream('pretreatment_steam',
                    phase='g',
                    T=268+273.15,
                    P=13*101325,
                    Water=24534+3490,
                    units='kg/hr')
    air = bst.Stream('air', O2=0.23, N2=0.77, phase='g', units='kg/hr')
    
    ### Pretreatment system
    ideal = ammonia.thermo.ideal()
    T201 = bst.StorageTank('T201', ammonia, tau=7, thermo=ideal)
    T201.ammonia_loading = ammonia_loading
    recycle = bst.Stream()
    P = pretreatment_steam.chemicals['H2O'].Psat(T_pretreatment_reactor + 25)
    M202 = bst.SteamMixer('M202', (feedstock, pretreatment_steam, warm_process_water, recycle), 
                          liquid_IDs=['Water', 'NH3'], P=P, solids_loading=solids_loading, thermo=ideal)
    R201 = bst.PretreatmentReactorSystem('R201', M202-0, tau=residence_time,
                                           T=T_pretreatment_reactor, thermo=ideal,
                                           reactions=pretreatment_reactions,
                                           run_vle=False)
    P201 = bst.BlowdownDischargePump('P201', R201-1, thermo=ideal)
    M205 = bst.Mixer('M205', (P201-0, air))
    F201 = bst.Flash('F201', M205-0, P=101325, T=310, thermo=ideal)
    @M205.add_specification(run=True)
    def update_air():
        feed, air = M205.ins
        flow = 100 * feed.F_vol
        air.imol['O2', 'N2'] = [flow * 0.23, flow * 0.77] # Assume equal volumes is enough
    
    M204 = bst.Mixer('M204', (R201-0, F201-0), thermo=ideal)
    F202 = bst.Flash('F202', M204-0, T=260., P=101325, thermo=ideal)
    F202.HXN_ignore = True
    @F202.add_specification
    def complete_recovery():
        feed = F202.ins[0]
        vap, liq = F202.outs
        vap.phase = 'g'
        ms = F202._multi_stream
        liq.mol = feed.mol
        ms.T = vap.T = liq.T = F202.T
        ms.P = vap.P = liq.P = F202.P
        vap.imol['N2', 'O2'], liq.imol['N2', 'O2'] = liq.imol['N2', 'O2'], 0.
        ms['g'].copy_flow(vap)
        ms['l'].copy_flow(liq)
        
    P203 = bst.Pump('P203', F202-1, P=10*101325)
    M206 = bst.Mixer('M206', (P203-0, T201-0), recycle)
    
    @M206.add_specification(run=True)
    def adjust_ammonia():
        recycle = M206.ins[0]
        fresh_ammonia = T201.ins[0]
        fresh_ammonia.imass['NH3'] = NH3_loss = F202.outs[0].imass['NH3'] + F201.outs[1].imass['NH3']
        required_ammonia = T201.ammonia_loading * (feedstock.F_mass - feedstock.imass['Water'])
        recycle.imass['NH3'] = required_ammonia - NH3_loss
        for i in T201.path_until(M206): i.run()
    
    if neutralize:
        sulfuric_acid = Stream(**s.sulfuric_acid)
        P202 = bst.HydrolyzatePump('P202', F201-1, thermo=ideal)
        H2SO4_storage = bst.SulfuricAcidStorageTank('H2SO4_storage', sulfuric_acid)
        T202 = bst.SulfuricAcidTank('T202', H2SO4_storage-0)
        M207 = bst.Mixer('M207', (T202-0, P202-0), pretreated_biomass)
        M207.neutralization_rxn = rxn.Rxn('2 NH3 + H2SO4 -> (NH4)2SO4', 'H2SO4', 1)
        @M207.add_specification
        def update_sulfuric_acid_loading():
            _, feed = M207.ins
            fresh_sulfuric_acid = H2SO4_storage.ins[0]
            fresh_sulfuric_acid.imol['H2SO4'] = feed.imol['NH3'] / 2
            for i in H2SO4_storage.path_until(M207): i.run()
            M207._run()
            M207.neutralization_rxn(M207.outs[0])
    else:
        P202 = bst.HydrolyzatePump('P202', F201-1, pretreated_biomass, thermo=ideal)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import os, sys, io
os.environ['NUMPY_EXPERIMENTAL_ARRAY_FUNCTION'] = '0'

import numpy as np
import cProfile, pstats
from pstats import SortKey

path = os.path.dirname(__file__)

# Instead using `pstats`, for t = 0.1, the stats are stored in `system_01`
cProfile.run('from biorefineries import lactic as la; la.load()', 'la_profile_temp.prof')

p = pstats.Stats('la_profile_temp.prof')

# Save all results to a csv file
output = io.StringIO()
# Save the old stream so we can swap back
console_strm = p.stream
p.stream = output

# Pass the outputs to `text`
p.sort_stats('tottime').print_stats()

# Add deliminator
text = output.getvalue()
text = 'ncalls' + text.split('ncalls')[-1]
text = '\n'.join([','.join(line.rstrip().split(None,5)) for line in text.split('\n')])

# Save the results to csv, here
csv_path = os.path.join(path, 'la_profile_temp.csv')
with open(csv_path, 'w+') as f:
    f.write(text)
    f.close()

# Swap back the output stream
p.stream = console_strm

# import biosteam as bst
# from chaospy import distributions as shape
# from biorefineries import cornstover as cs
# cs.load()

# model = bst.Model(cs.cornstover_sys, exception_hook='raise')

# param = model.parameter

# D = shape.Uniform(4, 6)
# @param(name='setter1', element=cs.cornstover, kind='coupled', distribution=D)
# def good_param_setter1(i):
#     cs.cornstover.price = i
#     print(f'Feedstock: {cs.cornstover.price}')

# D = shape.Uniform(4, 6)
# @param(name='setter2', element=cs.cornstover, kind='coupled', distribution=D)
# def bad_param_setter(i):
#     assert False # this clear won't work

# D = shape.Uniform(4, 6)
# @param(name='setter3', element=cs.ethanol, kind='coupled', distribution=D)
# def good_param_setter2(i):
#     cs.ethanol.price = i
#     print(f'Ethanol: {cs.ethanol.price}')


# model.metric(lambda: cs.cornstover.price, 'Cornstover price')
# model.metric(lambda: cs.ethanol.price, 'Ethanol price')

# samples = model.sample(N=2, rule='L')
# model.load_samples(samples)
# model.evaluate()


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""
from biosteam import stream_kwargs

cornstover = stream_kwargs(
    'cornstover',
    Glucan=0.28,
    Xylan=0.1562,
    Galactan=0.001144,
    Arabinan=0.01904,
    Mannan=0.0048,
    Lignin=0.12608,
    Acetate=0.01448,
    Protein=0.0248,
    Extract=0.1172,
    Ash=0.03944,
    Sucrose=0.00616,
    Water=0.2,
    total_flow=104229.16,
    units='kg/hr',
    price=0.05158816935126135,
)

switchgrass = stream_kwargs(
    'switchgrass',
    Arabinan=0.02789023841655421,
    Galactan=0.010436347278452543,
    Glucan=0.2717049032838507,
    Xylan=0.21214574898785432,
    Mannan=0.005937921727395412,
    Lignin=0.17112010796221322,
    Ash=0.016194331983805668,
    Extractives=0.08457040035987407,
    Water=0.2,
    total_flow=104229.16,
    units='kg/hr',
    price=0.08, # Price of switchgrass, table 4, Madhu Khanna et al. (Costs of producing miscanthus and switchgrass for bioenergy in Illinois);  https://www.sciencedirect.com/science/article/pii/S096195340700205X?casa_token=KfYfzJtDwv0AAAAA:OqeJmpofk1kIgFk2DcUvXNG35qYwlWvPKZ7ENI3R6RUKeoahiTDpOhhd_mpLtRthTGuXJKDzMOc
)

NaOH = stream_kwargs(
    'NaOH', 
    NaOH=1, 
    price=0.14952852932689323
)

ammonia = stream_kwargs(
    'ammonia', 
    phase='l',
    NH3=1, 
    P=12 * 101325, 
    price=0.4485966110937889,
)

sulfuric_acid = stream_kwargs(
    'sulfuric_acid',
    P=5.4*101325,
    T=294.15,
    Water=130,
    H2SO4=1800,
    units='kg/hr',
    price=0.0897171175961359
)
denaturant = stream_kwargs(
    'denaturant',
    Octane=1,
    price=0.756,
)
ethanol = stream_kwargs(
    'ethanol',
    price=0.7198608114634679,
)
lignin = stream_kwargs('lignin')
slurry = stream_kwargs('slurry')
DAP = stream_kwargs('DAP', price=0.9869213628968231)
CSL = stream_kwargs('CSL', price=0.0568241480781522)
vent = stream_kwargs('vent')
beer = stream_kwargs('beer')
pretreated_biomass = stream_kwargs('pretreated_biomass')
cellulase = stream_kwargs('cellulase', price=0.212)
saccharification_water = stream_kwargs('saccharification_water')
slurry = stream_kwargs('slurry')
nanofilter_retentate = stream_kwargs('nanofilter_retentate')
pretreatment_wastewater = stream_kwargs('pretreatment_wastewater')
juice = stream_kwargs('juice')
hydrolysate = stream_kwargs('hydrolysate')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
from biosteam import SystemFactory
from .. import streams as s

__all__ = (
    'create_bagasse_drying_system',
    'create_bagasse_pelleting_system',
)


@SystemFactory(
    ID='bagasse_pelleting_sys',
    ins=[s.bagasse],
    outs=[s.bagasse_pellets]
)
def create_bagasse_pelleting_system(ins, outs):
    bagasse, = ins
    bagasse_pellets, = outs
    U401 = bst.HammerMill('U401', bagasse)
    U402 = bst.DrumDryer('U402', 
        (U401-0, 'dryer_air', 'dryer_natural_gas'), 
        ('', 'dryer_outlet_air', 'dryer_emissions'),
        moisture_content=0.18, split=0.,
        utility_agent='Steam',
    )
    # X401 = bst.ThermalOxidizer('X401', (U403-1, 'oxidizer_air'), 'oxidizer_emissions')
    U403 = bst.ScrewFeeder('U403', U402-0)
    U404 = bst.BagassePelletMill('U404', U403-0)
    U405 = bst.ConveyingBelt('U405', U404-0, bagasse_pellets)

@SystemFactory(
    ID='bagasse_drying_sys',
    ins=[s.bagasse],
    outs=['dried_bagasse']
)
def create_bagasse_drying_system(ins, outs, utility_agent='Steam'):
    U402 = bst.DrumDryer('U402', 
        (ins[0], 'dryer_air', 'dryer_natural_gas'), 
        ('', 'dryer_outlet_air', 'dryer_emissions'),
        moisture_content=0.40, split=0.,
        utility_agent=utility_agent,
    )
    bst.ConveyingBelt('U403', U402-0, outs)
